#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from typing import Any, Callable, Dict, List, Optional, Tuple

import numpy as np
import torch
from ax.models.base import Model
from ax.models.model_utils import tunable_feature_indices
from ax.models.random.base import RandomModel
from ax.models.types import TConfig
from ax.utils.common.docutils import copy_doc
from ax.utils.common.typeutils import not_none
from torch.quasirandom import SobolEngine


class SobolGenerator(RandomModel):
    """This class specifies the generation algorithm for a Sobol generator.

    As Sobol does not make use of a model, it does not implement
    the fit or predict methods.

    Attributes:
        init_position: The initial state of the Sobol generator.
            Starts at 0 by default.
        scramble: If True, permutes the parameter values among
            the elements of the Sobol sequence. Default is True.
        See base `RandomModel` for a description of remaining attributes.
    """

    def __init__(
        self,
        seed: Optional[int] = None,
        deduplicate: bool = True,
        init_position: int = 0,
        scramble: bool = True,
        generated_points: Optional[np.ndarray] = None,
        fallback_to_sample_polytope: bool = False,
    ) -> None:
        super().__init__(
            deduplicate=deduplicate,
            seed=seed,
            generated_points=generated_points,
            fallback_to_sample_polytope=fallback_to_sample_polytope,
        )
        self.init_position = init_position
        self.scramble = scramble
        # Initialize engine on gen.
        self._engine: Optional[SobolEngine] = None

    def init_engine(self, n_tunable_features: int) -> SobolEngine:
        """Initialize singleton SobolEngine, only on gen.

        Args:
            n_tunable_features: The number of features which can be
                searched over.

        Returns:
            SobolEngine, which can generate Sobol points.

        """
        if not self._engine:
            self._engine = SobolEngine(
                dimension=n_tunable_features, scramble=self.scramble, seed=self.seed
            ).fast_forward(self.init_position)
        return self._engine

    @property
    def engine(self) -> Optional[SobolEngine]:
        """Return a singleton SobolEngine."""
        return self._engine

    def gen(
        self,
        n: int,
        bounds: List[Tuple[float, float]],
        linear_constraints: Optional[Tuple[np.ndarray, np.ndarray]] = None,
        fixed_features: Optional[Dict[int, float]] = None,
        model_gen_options: Optional[TConfig] = None,
        rounding_func: Optional[Callable[[np.ndarray], np.ndarray]] = None,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Generate new candidates.

        Args:
            n: Number of candidates to generate.
            bounds: A list of (lower, upper) tuples for each column of X.
            linear_constraints: A tuple of (A, b). For k linear constraints on
                d-dimensional x, A is (k x d) and b is (k x 1) such that
                A x <= b.
            fixed_features: A map {feature_index: value} for features that
                should be fixed to a particular value during generation.
            rounding_func: A function that rounds an optimization result
                appropriately (e.g., according to `round-trip` transformations).

        Returns:
            2-element tuple containing

            - (n x d) array of generated points.
            - Uniform weights, an n-array of ones for each point.

        """
        tf_indices = tunable_feature_indices(
            bounds=bounds, fixed_features=fixed_features
        )
        if len(tf_indices) > 0:
            self.init_engine(len(tf_indices))
        points, weights = super().gen(
            n=n,
            bounds=bounds,
            linear_constraints=linear_constraints,
            fixed_features=fixed_features,
            model_gen_options=model_gen_options,
            rounding_func=rounding_func,
        )
        if self.engine:
            self.init_position = not_none(self.engine).num_generated
        return (points, weights)

    @copy_doc(Model._get_state)
    def _get_state(self) -> Dict[str, Any]:
        state = super()._get_state()
        state.update({"init_position": self.init_position})
        return state

    def _gen_samples(self, n: int, tunable_d: int) -> np.ndarray:
        """Generate n samples.

        Args:
            n: Number of samples to generate.
            tunable_d: The dimension of the generated samples. This must
                match the tunable parameters used while initializing the
                Sobol engine.

        Returns:
            A numpy array of samples of shape `(n x tunable_d)`.
        """
        if tunable_d == 0:
            return np.zeros((n, 0))
        if self.engine is None:
            raise ValueError(
                "Sobol Engine must be initialized before candidate generation."
            )
        return not_none(self.engine).draw(n, dtype=torch.double).numpy()



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import json
import time
import warnings
from abc import ABC
from collections import OrderedDict
from copy import deepcopy
from dataclasses import dataclass, field

from logging import Logger
from typing import Any, Dict, List, MutableMapping, Optional, Set, Tuple, Type

from ax.core.arm import Arm
from ax.core.base_trial import NON_ABANDONED_STATUSES, TrialStatus
from ax.core.data import Data
from ax.core.experiment import Experiment
from ax.core.generator_run import extract_arm_predictions, GeneratorRun
from ax.core.observation import (
    Observation,
    ObservationData,
    ObservationFeatures,
    observations_from_data,
    recombine_observations,
    separate_observations,
)
from ax.core.optimization_config import OptimizationConfig
from ax.core.parameter import ParameterType, RangeParameter
from ax.core.search_space import SearchSpace
from ax.core.types import TCandidateMetadata, TModelCov, TModelMean, TModelPredict
from ax.exceptions.core import UnsupportedError, UserInputError
from ax.modelbridge.transforms.base import Transform
from ax.modelbridge.transforms.cast import Cast
from ax.models.types import TConfig
from ax.utils.common.logger import get_logger
from ax.utils.common.typeutils import checked_cast, not_none
from botorch.exceptions.warnings import InputDataWarning

logger: Logger = get_logger(__name__)


@dataclass(frozen=True)
class BaseGenArgs:
    search_space: SearchSpace
    optimization_config: Optional[OptimizationConfig]
    pending_observations: Dict[str, List[ObservationFeatures]]
    fixed_features: Optional[ObservationFeatures]


@dataclass(frozen=True)
class GenResults:
    observation_features: List[ObservationFeatures]
    weights: List[float]
    best_observation_features: Optional[ObservationFeatures] = None
    gen_metadata: Dict[str, Any] = field(default_factory=dict)


class ModelBridge(ABC):  # noqa: B024 -- ModelBridge doesn't have any abstract methods.
    """The main object for using models in Ax.

    ModelBridge specifies 3 methods for using models:

    - predict: Make model predictions. This method is not optimized for
      speed and so should be used primarily for plotting or similar tasks
      and not inside an optimization loop.
    - gen: Use the model to generate new candidates.
    - cross_validate: Do cross validation to assess model predictions.

    ModelBridge converts Ax types like Data and Arm to types that are
    meant to be consumed by the models. The data sent to the model will depend
    on the implementation of the subclass, which will specify the actual API
    for external model.

    This class also applies a sequence of transforms to the input data and
    problem specification which can be used to ensure that the external model
    receives appropriate inputs.

    Subclasses will implement what is here referred to as the "terminal
    transform," which is a transform that changes types of the data and problem
    specification.
    """

    def __init__(
        self,
        search_space: SearchSpace,
        # pyre-fixme[2]: Parameter annotation cannot be `Any`.
        model: Any,
        transforms: Optional[List[Type[Transform]]] = None,
        experiment: Optional[Experiment] = None,
        data: Optional[Data] = None,
        transform_configs: Optional[Dict[str, TConfig]] = None,
        status_quo_name: Optional[str] = None,
        status_quo_features: Optional[ObservationFeatures] = None,
        optimization_config: Optional[OptimizationConfig] = None,
        fit_out_of_design: bool = False,
        fit_abandoned: bool = False,
        fit_tracking_metrics: bool = True,
        fit_on_init: bool = True,
    ) -> None:
        """
        Applies transforms and fits model.

        Args:
            experiment: Is used to get arm parameters. Is not mutated.
            search_space: Search space for fitting the model. Constraints need
                not be the same ones used in gen.
            data: Ax Data.
            model: Interface will be specified in subclass. If model requires
                initialization, that should be done prior to its use here.
            transforms: List of uninitialized transform classes. Forward
                transforms will be applied in this order, and untransforms in
                the reverse order.
            transform_configs: A dictionary from transform name to the
                transform config dictionary.
            status_quo_name: Name of the status quo arm. Can only be used if
                Data has a single set of ObservationFeatures corresponding to
                that arm.
            status_quo_features: ObservationFeatures to use as status quo.
                Either this or status_quo_name should be specified, not both.
            optimization_config: Optimization config defining how to optimize
                the model.
            fit_out_of_design: If specified, all training data is returned.
                Otherwise, only in design points are returned.
            fit_abandoned: Whether data for abandoned arms or trials should be
                included in model training data. If ``False``, only
                non-abandoned points are returned.
            fit_tracking_metrics: Whether to fit a model for tracking metrics.
                Setting this to False will improve runtime at the expense of
                models not being available for predicting tracking metrics.
                NOTE: This can only be set to False when the optimization config
                is provided.
            fit_on_init: Whether to fit the model on initialization. This can
                be used to skip model fitting when a fitted model is not needed.
                To fit the model afterwards, use `_process_and_transform_data`
                to get the transformed inputs and call `_fit_if_implemented` with
                the transformed inputs.
        """
        t_fit_start = time.monotonic()
        transforms = transforms or []
        # pyre-ignore: Cast is a Tranform
        transforms: List[Type[Transform]] = [Cast] + transforms

        self.fit_time: float = 0.0
        self.fit_time_since_gen: float = 0.0
        self._metric_names: Set[str] = set()
        self._training_data: List[Observation] = []
        self._optimization_config: Optional[OptimizationConfig] = optimization_config
        self._training_in_design: List[bool] = []
        self._status_quo: Optional[Observation] = None
        self._status_quo_name: Optional[str] = None
        self._arms_by_signature: Optional[Dict[str, Arm]] = None
        self.transforms: MutableMapping[str, Transform] = OrderedDict()
        self._model_key: Optional[str] = None
        self._model_kwargs: Optional[Dict[str, Any]] = None
        self._bridge_kwargs: Optional[Dict[str, Any]] = None
        self._model_space: SearchSpace = search_space.clone()
        self._raw_transforms = transforms
        self._transform_configs: Optional[Dict[str, TConfig]] = transform_configs
        self._fit_out_of_design = fit_out_of_design
        self._fit_abandoned = fit_abandoned
        self._fit_tracking_metrics = fit_tracking_metrics
        self.outcomes: List[str] = []
        self._experiment_has_immutable_search_space_and_opt_config: bool = (
            experiment is not None and experiment.immutable_search_space_and_opt_config
        )
        self._experiment_properties: Dict[str, Any] = {}

        if experiment is not None:
            if self._optimization_config is None:
                self._optimization_config = experiment.optimization_config
            self._arms_by_signature = experiment.arms_by_signature
            self._experiment_properties = experiment._properties

        if self._fit_tracking_metrics is False:
            if self._optimization_config is None:
                raise UserInputError(
                    "Optimization config is required when "
                    "`fit_tracking_metrics` is False."
                )
            self.outcomes = sorted(self._optimization_config.metrics.keys())

        # Set training data (in the raw / untransformed space). This also omits
        # out-of-design and abandoned observations depending on the corresponding flags.
        observations_raw = self._prepare_observations(experiment=experiment, data=data)
        observations_raw = self._set_training_data(
            observations=observations_raw, search_space=self._model_space
        )

        # Set model status quo.
        # NOTE: training data must be set before setting the status quo.
        self._set_status_quo(
            experiment=experiment,
            status_quo_name=status_quo_name,
            status_quo_features=status_quo_features,
        )

        # Save model, apply terminal transform, and fit.
        self.model = model
        if fit_on_init:
            observations, search_space = self._transform_data(
                observations=observations_raw,
                search_space=self._model_space,
                transforms=self._raw_transforms,
                transform_configs=self._transform_configs,
            )
            self._fit_if_implemented(
                search_space=search_space,
                observations=observations,
                time_so_far=time.monotonic() - t_fit_start,
            )

    def _fit_if_implemented(
        self,
        search_space: SearchSpace,
        observations: List[Observation],
        time_so_far: float,
    ) -> None:
        r"""Fits the model if `_fit` is implemented and stores fit time.

        Args:
            search_space: A transformed search space for fitting the model.
            observations: The observations to fit the model with. These should
                also be transformed.
            time_so_far: Time spent in initializing the model up to
                `_fit_if_implemented` call.
        """
        try:
            t_fit_start = time.monotonic()
            self._fit(
                model=self.model,
                search_space=search_space,
                observations=observations,
            )
            increment = time.monotonic() - t_fit_start + time_so_far
            self.fit_time += increment
            self.fit_time_since_gen += increment
        except NotImplementedError:
            pass

    def _process_and_transform_data(
        self,
        experiment: Optional[Experiment] = None,
        data: Optional[Data] = None,
    ) -> Tuple[List[Observation], SearchSpace]:
        r"""Processes the data into observations and returns transformed
        observations and the search space. This packages the following methods:
        * self._prepare_observations
        * self._set_training_data
        * self._transform_data
        """
        observations = self._prepare_observations(experiment=experiment, data=data)
        observations_raw = self._set_training_data(
            observations=observations, search_space=self._model_space
        )
        return self._transform_data(
            observations=observations_raw,
            search_space=self._model_space,
            transforms=self._raw_transforms,
            transform_configs=self._transform_configs,
        )

    def _prepare_observations(
        self, experiment: Optional[Experiment], data: Optional[Data]
    ) -> List[Observation]:
        if experiment is None or data is None:
            return []
        return observations_from_data(
            experiment=experiment,
            data=data,
            statuses_to_include=self.statuses_to_fit,
            statuses_to_include_map_metric=self.statuses_to_fit_map_metric,
        )

    def _transform_data(
        self,
        observations: List[Observation],
        search_space: SearchSpace,
        transforms: Optional[List[Type[Transform]]],
        transform_configs: Optional[Dict[str, TConfig]],
        assign_transforms: bool = True,
    ) -> Tuple[List[Observation], SearchSpace]:
        """Initialize transforms and apply them to provided data."""
        # Initialize transforms
        search_space = search_space.clone()
        if transforms is not None:
            if transform_configs is None:
                transform_configs = {}

            for t in transforms:
                t_instance = t(
                    search_space=search_space,
                    observations=observations,
                    modelbridge=self,
                    config=transform_configs.get(t.__name__, None),
                )
                search_space = t_instance.transform_search_space(search_space)
                observations = t_instance.transform_observations(observations)
                if assign_transforms:
                    self.transforms[t.__name__] = t_instance

        return observations, search_space

    def _prepare_training_data(
        self, observations: List[Observation]
    ) -> List[Observation]:
        observation_features, observation_data = separate_observations(observations)
        if len(observation_features) != len(set(observation_features)):
            raise ValueError(
                "Observation features are not unique. "
                "Something went wrong constructing training data..."
            )
        return observations

    def _set_training_data(
        self, observations: List[Observation], search_space: SearchSpace
    ) -> List[Observation]:
        """Store training data, not-transformed.

        If the modelbridge specifies _fit_out_of_design, all training data is
        returned. Otherwise, only in design points are returned.
        """
        observations = self._prepare_training_data(observations=observations)
        self._training_data = deepcopy(observations)
        self._metric_names: Set[str] = set()
        for obs in observations:
            self._metric_names.update(obs.data.metric_names)
        return self._process_in_design(
            search_space=search_space,
            observations=observations,
        )

    def _extend_training_data(
        self, observations: List[Observation]
    ) -> List[Observation]:
        """Extend and return training data, not-transformed.

        If the modelbridge specifies _fit_out_of_design, all training data is
        returned. Otherwise, only in design points are returned.

        Args:
            observations: New observations.

        Returns: New + old observations.
        """
        observations = self._prepare_training_data(observations=observations)
        for obs in observations:
            for metric_name in obs.data.metric_names:
                if metric_name not in self._metric_names:
                    raise ValueError(
                        f"Unrecognised metric {metric_name}; cannot update "
                        "training data with metrics that were not in the original "
                        "training data."
                    )
        # Initialize with all points in design.
        self._training_data.extend(deepcopy(observations))
        all_observations = self.get_training_data()
        return self._process_in_design(
            search_space=self._model_space,
            observations=all_observations,
        )

    def _process_in_design(
        self,
        search_space: SearchSpace,
        observations: List[Observation],
    ) -> List[Observation]:
        """Set training_in_design, and decide whether to filter out of design points."""
        # Don't filter points.
        if self._fit_out_of_design:
            # Use all data for training
            # Set training_in_design to True for all observations so that
            # all observations are used in CV and plotting
            self.training_in_design = [True] * len(observations)
            return observations
        in_design = self._compute_in_design(
            search_space=search_space, observations=observations
        )
        self.training_in_design = in_design
        in_design_obs = [
            observations[i] for i, is_in_design in enumerate(in_design) if is_in_design
        ]
        return in_design_obs

    def _compute_in_design(
        self,
        search_space: SearchSpace,
        observations: List[Observation],
    ) -> List[bool]:
        return [
            search_space.check_membership(obs.features.parameters)
            for obs in observations
        ]

    def _set_status_quo(
        self,
        experiment: Optional[Experiment],
        status_quo_name: Optional[str],
        status_quo_features: Optional[ObservationFeatures],
    ) -> None:
        """Set model status quo by matching status_quo_name or status_quo_features.

        First checks for status quo in inputs status_quo_name and
        status_quo_features. If neither of these is provided, checks the
        experiment for a status quo. If that is set, it is handled by name in
        the same way as input status_quo_name.

        Args:
            experiment: Experiment that will be checked for status quo.
            status_quo_name: Name of status quo arm.
            status_quo_features: Features for status quo.
        """
        self._status_quo: Optional[Observation] = None
        sq_obs = None

        if (
            status_quo_name is None
            and status_quo_features is None
            and experiment is not None
            and experiment.status_quo is not None
        ):
            status_quo_name = experiment.status_quo.name

        if status_quo_name is not None:
            if status_quo_features is not None:
                raise ValueError(
                    "Specify either status_quo_name or status_quo_features, not both."
                )
            sq_obs = [
                obs for obs in self._training_data if obs.arm_name == status_quo_name
            ]
        elif status_quo_features is not None:
            sq_obs = [
                obs
                for obs in self._training_data
                if (obs.features.parameters == status_quo_features.parameters)
                and (obs.features.trial_index == status_quo_features.trial_index)
            ]

        # if status_quo_name or status_quo_features is used for matching status quo
        if sq_obs is not None:
            if len(sq_obs) == 0:
                logger.warning(f"Status quo {status_quo_name} not present in data")
            elif len(sq_obs) >= 1:
                # status quo name (not features as trial index is part of the
                # observation features) should be consistent even if we have multiple
                # observations of the status quo.
                # This is useful for getting status_quo_data_by_trial
                self._status_quo_name = sq_obs[0].arm_name
                if len(sq_obs) > 1:
                    logger.warning(
                        f"Status quo {status_quo_name} found in data with multiple "
                        "features. Use status_quo_features to specify which to use."
                    )
                else:
                    # if there is a unique status_quo, set it
                    # unique features verified in _set_training_data.
                    self._status_quo = sq_obs[0]

    @property
    def status_quo_data_by_trial(self) -> Optional[Dict[int, ObservationData]]:
        """A map of trial index to the status quo observation data of each trial"""
        return _get_status_quo_by_trial(
            observations=self._training_data,
            status_quo_name=(
                self._status_quo_name
                if self.status_quo is None
                else self.status_quo.arm_name
            ),
            status_quo_features=(
                None if self.status_quo is None else self.status_quo.features
            ),
        )

    @property
    def status_quo(self) -> Optional[Observation]:
        """Observation corresponding to status quo, if any."""
        return self._status_quo

    @property
    def metric_names(self) -> Set[str]:
        """Metric names present in training data."""
        return self._metric_names

    @property
    def model_space(self) -> SearchSpace:
        """SearchSpace used to fit model."""
        return self._model_space

    def get_training_data(self) -> List[Observation]:
        """A copy of the (untransformed) data with which the model was fit."""
        return deepcopy(self._training_data)

    @property
    def training_in_design(self) -> List[bool]:
        """For each observation in the training data, a bool indicating if it
        is in-design for the model.
        """
        return self._training_in_design

    @property
    def statuses_to_fit(self) -> Set[TrialStatus]:
        """Statuses to fit the model on."""
        if self._fit_abandoned:
            return set(TrialStatus)
        return NON_ABANDONED_STATUSES

    @property
    def statuses_to_fit_map_metric(self) -> Set[TrialStatus]:
        """Statuses to fit the model on."""
        return {TrialStatus.COMPLETED}

    @training_in_design.setter
    def training_in_design(self, training_in_design: List[bool]) -> None:
        if len(training_in_design) != len(self._training_data):
            raise ValueError(
                f"In-design indicators not same length ({len(training_in_design)})"
                f" as training data ({len(self._training_data)})."
            )
        # Identify out-of-design arms
        if sum(training_in_design) < len(training_in_design):
            ood_names = []
            for i, obs in enumerate(self._training_data):
                if not training_in_design[i] and obs.arm_name is not None:
                    ood_names.append(obs.arm_name)
            ood_str = ", ".join(set(ood_names))
            logger.info(f"Leaving out out-of-design observations for arms: {ood_str}")
        self._training_in_design = training_in_design

    def _fit(
        self,
        # pyre-fixme[2]: Parameter annotation cannot be `Any`.
        model: Any,
        search_space: SearchSpace,
        observations: List[Observation],
    ) -> None:
        """Apply terminal transform and fit model."""
        raise NotImplementedError

    def _batch_predict(
        self, observation_features: List[ObservationFeatures]
    ) -> List[ObservationData]:
        """Predict a list of ObservationFeatures together."""
        # Get modifiable version
        observation_features = deepcopy(observation_features)

        # Transform
        for t in self.transforms.values():
            observation_features = t.transform_observation_features(
                observation_features
            )
        # Apply terminal transform and predict
        observation_data = self._predict(observation_features)

        # Apply reverse transforms, in reverse order
        pred_observations = recombine_observations(
            observation_features=observation_features, observation_data=observation_data
        )

        for t in reversed(list(self.transforms.values())):
            pred_observations = t.untransform_observations(pred_observations)
        return [obs.data for obs in pred_observations]

    def _single_predict(
        self, observation_features: List[ObservationFeatures]
    ) -> List[ObservationData]:
        """Predict one ObservationFeature at a time."""
        observation_data = []
        for obsf in observation_features:
            try:
                obsd = self._batch_predict([obsf])
                observation_data += obsd
            except (TypeError, ValueError) as e:
                # If the prediction is not out of design, this is a real error.
                # Let's re-raise.
                if self.model_space.check_membership(obsf.parameters):
                    logger.debug(obsf.parameters)
                    logger.debug(self.model_space)
                    raise e from None
                # Prediction is out of design.
                # Training data is untranformed already.
                observation = next(
                    (
                        data
                        for data in self.get_training_data()
                        if obsf.parameters == data.features.parameters
                        and obsf.trial_index == data.features.trial_index
                    ),
                    None,
                )
                if not observation:
                    raise ValueError(
                        "Out-of-design point could not be transformed, and was "
                        "not found in the training data."
                    )
                observation_data.append(observation.data)
        return observation_data

    def _predict_observation_data(
        self, observation_features: List[ObservationFeatures]
    ) -> List[ObservationData]:
        """
        Like 'predict' method, but returns results as a list of ObservationData

        Predictions are made for all outcomes.
        If an out-of-design observation can successfully be transformed,
        the predicted value will be returned.
        Othwerise, we will attempt to find that observation in the training data
        and return the raw value.

        Args:
            observation_features: observation features

        Returns:
            List of `ObservationData`
        """
        # Predict in single batch.
        try:
            observation_data = self._batch_predict(observation_features)
        # Predict one by one.
        except (TypeError, ValueError):
            observation_data = self._single_predict(observation_features)
        return observation_data

    def predict(self, observation_features: List[ObservationFeatures]) -> TModelPredict:
        """Make model predictions (mean and covariance) for the given
        observation features.

        Predictions are made for all outcomes.
        If an out-of-design observation can successfully be transformed,
        the predicted value will be returned.
        Othwerise, we will attempt to find that observation in the training data
        and return the raw value.

        Args:
            observation_features: observation features

        Returns:
            2-element tuple containing

            - Dictionary from metric name to list of mean estimates, in same
              order as observation_features.
            - Nested dictionary with cov['metric1']['metric2'] a list of
              cov(metric1@x, metric2@x) for x in observation_features.
        """
        observation_data = self._predict_observation_data(
            observation_features=observation_features
        )
        f, cov = unwrap_observation_data(observation_data)
        return f, cov

    def _predict(
        self, observation_features: List[ObservationFeatures]
    ) -> List[ObservationData]:
        """Apply terminal transform, predict, and reverse terminal transform on
        output.
        """
        raise NotImplementedError

    def update(self, new_data: Data, experiment: Experiment) -> None:
        """Update the model bridge and the underlying model with new data. This
        method should be used instead of `fit`, in cases where the underlying
        model does not need to be re-fit from scratch, but rather updated.

        Note: `update` expects only new data (obtained since the model initialization
        or last update) to be passed in, not all data in the experiment.

        Args:
            new_data: Data from the experiment obtained since the last call to
                `update`.
            experiment: Experiment, in which this data was obtained.
        """
        raise DeprecationWarning("ModelBridge.update is deprecated. Use `fit` instead.")

    def _get_transformed_gen_args(
        self,
        search_space: SearchSpace,
        optimization_config: Optional[OptimizationConfig] = None,
        pending_observations: Optional[Dict[str, List[ObservationFeatures]]] = None,
        fixed_features: Optional[ObservationFeatures] = None,
    ) -> BaseGenArgs:
        if pending_observations is None:
            pending_observations = {}
        if optimization_config is None:
            optimization_config = (
                self._optimization_config.clone()
                if self._optimization_config is not None
                else None
            )
        else:
            if not self._fit_tracking_metrics:
                # Check that the optimization config has the same metrics as
                # the original one. Otherwise, we may attempt to optimize over
                # metrics that do not have a fitted model.
                outcomes = set(optimization_config.metrics.keys())
                if not outcomes.issubset(self.outcomes):
                    raise UnsupportedError(
                        "When fit_tracking_metrics is False, the optimization config "
                        "can only include metrics that were included in the "
                        "optimization config used while initializing the ModelBridge. "
                        f"Metrics {outcomes} is not a subset of {self.outcomes}."
                    )
            optimization_config = optimization_config.clone()

        # TODO(T34225037): replace deepcopy with native clone() in Ax
        pending_observations = deepcopy(pending_observations)
        fixed_features = deepcopy(fixed_features)
        search_space = search_space.clone()

        # Transform
        for t in self.transforms.values():
            search_space = t.transform_search_space(search_space)
            if optimization_config is not None:
                optimization_config = t.transform_optimization_config(
                    optimization_config=optimization_config,
                    modelbridge=self,
                    fixed_features=fixed_features,
                )
            for metric, po in pending_observations.items():
                pending_observations[metric] = t.transform_observation_features(po)
            fixed_features = (
                t.transform_observation_features([fixed_features])[0]
                if fixed_features is not None
                else None
            )
        return BaseGenArgs(
            search_space=search_space,
            optimization_config=optimization_config,
            pending_observations=pending_observations,
            fixed_features=fixed_features,
        )

    def _validate_gen_inputs(
        self,
        n: int,
        search_space: Optional[SearchSpace] = None,
        optimization_config: Optional[OptimizationConfig] = None,
        pending_observations: Optional[Dict[str, List[ObservationFeatures]]] = None,
        fixed_features: Optional[ObservationFeatures] = None,
        model_gen_options: Optional[TConfig] = None,
    ) -> None:
        """Validate inputs to `ModelBridge.gen`.

        Currently, this is only used to ensure that `n` is a positive integer.
        """
        if n < 1:
            raise UserInputError(
                f"Attempted to generate n={n} points. Number of points to generate "
                "must be a positive integer."
            )

    def gen(
        self,
        n: int,
        search_space: Optional[SearchSpace] = None,
        optimization_config: Optional[OptimizationConfig] = None,
        pending_observations: Optional[Dict[str, List[ObservationFeatures]]] = None,
        fixed_features: Optional[ObservationFeatures] = None,
        model_gen_options: Optional[TConfig] = None,
    ) -> GeneratorRun:
        """
        Generate new points from the underlying model according to
        search_space, optimization_config and other parameters.

        Args:
            n: Number of points to generate
            search_space: Search space
            optimization_config: Optimization config
            pending_observations: A map from metric name to pending
                observations for that metric.
            fixed_features: An ObservationFeatures object containing any
                features that should be fixed at specified values during
                generation.
            model_gen_options: A config dictionary that is passed along to the
                model. See `TorchOptConfig` for details.

        Returns:
            A GeneratorRun object that contains the generated points and other metadata.
        """
        self._validate_gen_inputs(
            n=n,
            search_space=search_space,
            optimization_config=optimization_config,
            pending_observations=pending_observations,
            fixed_features=fixed_features,
            model_gen_options=model_gen_options,
        )
        t_gen_start = time.monotonic()
        # Get modifiable versions
        if search_space is None:
            search_space = self._model_space
        orig_search_space = search_space
        search_space = search_space.clone()
        base_gen_args = self._get_transformed_gen_args(
            search_space=search_space,
            optimization_config=optimization_config,
            pending_observations=pending_observations,
            fixed_features=fixed_features,
        )
        # Apply terminal transform and gen
        gen_results = self._gen(
            n=n,
            search_space=base_gen_args.search_space,
            optimization_config=base_gen_args.optimization_config,
            pending_observations=base_gen_args.pending_observations,
            fixed_features=base_gen_args.fixed_features,
            model_gen_options=model_gen_options,
        )

        observation_features = gen_results.observation_features
        best_obsf = gen_results.best_observation_features

        # Apply reverse transforms
        for t in reversed(list(self.transforms.values())):
            observation_features = t.untransform_observation_features(
                observation_features
            )
            if best_obsf is not None:
                best_obsf = t.untransform_observation_features([best_obsf])[0]

        # Clamp the untransformed data to the original search space if
        # we don't fit/gen OOD points
        if not self._fit_out_of_design:
            observation_features = clamp_observation_features(
                observation_features, orig_search_space
            )
            if best_obsf is not None:
                best_obsf = clamp_observation_features([best_obsf], orig_search_space)[
                    0
                ]
        best_point_predictions = None
        try:
            model_predictions = self.predict(observation_features)
            if best_obsf is not None:
                best_point_predictions = extract_arm_predictions(
                    model_predictions=self.predict([best_obsf]), arm_idx=0
                )
        except NotImplementedError:
            model_predictions = None

        if best_obsf is None:
            best_arm = None
        else:
            best_arms, _ = gen_arms(
                observation_features=[best_obsf],
                arms_by_signature=self._arms_by_signature,
            )
            best_arm = best_arms[0]

        arms, candidate_metadata = gen_arms(
            observation_features=observation_features,
            arms_by_signature=self._arms_by_signature,
        )
        # If experiment has immutable search space and metrics, no need to
        # save them on generator runs.
        immutable = getattr(
            self, "_experiment_has_immutable_search_space_and_opt_config", False
        )
        optimization_config = None if immutable else base_gen_args.optimization_config
        gr = GeneratorRun(
            arms=arms,
            weights=gen_results.weights,
            optimization_config=optimization_config,
            search_space=None if immutable else base_gen_args.search_space,
            model_predictions=model_predictions,
            best_arm_predictions=(
                None if best_arm is None else (best_arm, best_point_predictions)
            ),
            fit_time=self.fit_time_since_gen,
            gen_time=time.monotonic() - t_gen_start,
            model_key=self._model_key,
            model_kwargs=self._model_kwargs,
            bridge_kwargs=self._bridge_kwargs,
            gen_metadata=gen_results.gen_metadata,
            model_state_after_gen=self._get_serialized_model_state(),
            candidate_metadata_by_arm_signature=candidate_metadata,
        )
        if len(gr.arms) < n:
            logger.warning(
                f"{self} was not able to generate {n} unique candidates. "
                "Generated arms have the following weights, as there are repeats:\n"
                f"{gr.weights}"
            )
        self.fit_time_since_gen = 0.0
        return gr

    def _gen(
        self,
        n: int,
        search_space: SearchSpace,
        optimization_config: Optional[OptimizationConfig],
        pending_observations: Dict[str, List[ObservationFeatures]],
        fixed_features: Optional[ObservationFeatures],
        model_gen_options: Optional[TConfig],
    ) -> GenResults:
        """Apply terminal transform, gen, and reverse terminal transform on
        output.
        """
        raise NotImplementedError

    def cross_validate(
        self,
        cv_training_data: List[Observation],
        cv_test_points: List[ObservationFeatures],
        use_posterior_predictive: bool = False,
    ) -> List[ObservationData]:
        """Make a set of cross-validation predictions.

        Args:
            cv_training_data: The training data to use for cross validation.
            cv_test_points: The test points at which predictions will be made.
            use_posterior_predictive: A boolean indicating if the predictions
                should be from the posterior predictive (i.e. including
                observation noise).

        Returns:
            A list of predictions at the test points.
        """
        # Apply transforms to cv_training_data and cv_test_points
        cv_training_data, cv_test_points, search_space = self._transform_inputs_for_cv(
            cv_training_data=cv_training_data, cv_test_points=cv_test_points
        )

        # Apply terminal transform, and get predictions.
        with warnings.catch_warnings():
            # Since each CV fold removes points from the training data, the remaining
            # observations will not pass the standardization test. To avoid confusing
            # users with this warning, we filter it out.
            warnings.filterwarnings(
                "ignore",
                message="Data is not standardized",
                category=InputDataWarning,
            )
            cv_predictions = self._cross_validate(
                search_space=search_space,
                cv_training_data=cv_training_data,
                cv_test_points=cv_test_points,
                use_posterior_predictive=use_posterior_predictive,
            )
        # Apply reverse transforms, in reverse order
        cv_test_observations = [
            Observation(features=obsf, data=cv_predictions[i])
            for i, obsf in enumerate(cv_test_points)
        ]

        for t in reversed(list(self.transforms.values())):
            cv_test_observations = t.untransform_observations(cv_test_observations)
        return [obs.data for obs in cv_test_observations]

    def _cross_validate(
        self,
        search_space: SearchSpace,
        cv_training_data: List[Observation],
        cv_test_points: List[ObservationFeatures],
        use_posterior_predictive: bool = False,
    ) -> List[ObservationData]:
        """Apply the terminal transform, make predictions on the test points,
        and reverse terminal transform on the results.
        """
        raise NotImplementedError

    def _transform_inputs_for_cv(
        self,
        cv_training_data: List[Observation],
        cv_test_points: List[ObservationFeatures],
    ) -> Tuple[List[Observation], List[ObservationFeatures], SearchSpace]:
        """Apply transforms to cv_training_data and cv_test_points,
        and return cv_training_data, cv_test_points, and search space in
        transformed space. This is to prepare data to be used in _cross_validate.

        Args:
            cv_training_data: The training data to use for cross validation.
            cv_test_points: The test points at which predictions will be made.

        Returns:
            cv_training_data, cv_test_points, and search space
            in transformed space."""
        cv_test_points = deepcopy(cv_test_points)
        cv_training_data = deepcopy(cv_training_data)
        search_space = self._model_space.clone()
        for t in self.transforms.values():
            cv_training_data = t.transform_observations(cv_training_data)
            cv_test_points = t.transform_observation_features(cv_test_points)
            search_space = t.transform_search_space(search_space)
        return cv_training_data, cv_test_points, search_space

    def _set_kwargs_to_save(
        self,
        model_key: str,
        model_kwargs: Dict[str, Any],
        bridge_kwargs: Dict[str, Any],
    ) -> None:
        """Set properties used to save the model that created a given generator
        run, on the `GeneratorRun` object. Each generator run produced by the
        `gen` method of this model bridge will have the model key and kwargs
        fields set as provided in arguments to this function.
        """
        self._model_key = model_key
        self._model_kwargs = model_kwargs
        self._bridge_kwargs = bridge_kwargs

    def _get_serialized_model_state(self) -> Dict[str, Any]:
        """Obtains the state of the underlying model (if using a stateful one)
        in a readily JSON-serializable form.
        """
        model = not_none(self.model)
        return model.serialize_state(raw_state=model._get_state())

    def _deserialize_model_state(
        self, serialized_state: Dict[str, Any]
    ) -> Dict[str, Any]:
        model = not_none(self.model)
        return model.deserialize_state(serialized_state=serialized_state)

    def feature_importances(self, metric_name: str) -> Dict[str, float]:
        """Computes feature importances for a single metric.

        Depending on the type of the model, this method will approach sensitivity
        analysis (calculating the sensitivity of the metric to changes in the search
        space's parameters, a.k.a. features) differently.

        For Bayesian optimization models (BoTorch models), this method uses parameter
        inverse lengthscales to compute normalized feature importances.

        NOTE: Currently, this is only implemented for GP models.

        Args:
            metric_name: Name of metric to compute feature importances for.

        Returns:
            A dictionary mapping parameter names to their corresponding feature
            importances.

        """
        raise NotImplementedError(
            "Feature importance not available for this model type"
        )

    # pyre-fixme[3]: Return annotation cannot be `Any`.
    def transform_observations(self, observations: List[Observation]) -> Any:
        """Applies transforms to given observation features and returns them in the
        model space.

        Args:
            observation_features: ObservationFeatures to be transformed.

        Returns:
            Transformed values. This could be e.g. a torch Tensor, depending
            on the ModelBridge subclass.
        """
        observations = deepcopy(observations)
        for t in self.transforms.values():
            observations = t.transform_observations(observations)
        # Apply terminal transform and return
        return self._transform_observations(observations)

    # pyre-fixme[3]: Return annotation cannot be `Any`.
    def _transform_observations(self, observations: List[Observation]) -> Any:
        """Apply terminal transform to given observations and return result."""
        raise NotImplementedError

    # pyre-fixme[3]: Return annotation cannot be `Any`.
    def transform_observation_features(
        self, observation_features: List[ObservationFeatures]
    ) -> Any:
        """Applies transforms to given observation features and returns them in the
        model space.

        Args:
            observation_features: ObservationFeatures to be transformed.

        Returns:
            Transformed values. This could be e.g. a torch Tensor, depending
            on the ModelBridge subclass.
        """
        obsf = deepcopy(observation_features)
        for t in self.transforms.values():
            obsf = t.transform_observation_features(obsf)
        # Apply terminal transform and return
        return self._transform_observation_features(obsf)

    # pyre-fixme[3]: Return annotation cannot be `Any`.
    def _transform_observation_features(
        self, observation_features: List[ObservationFeatures]
    ) -> Any:
        """Apply terminal transform to given observation features and return result."""
        raise NotImplementedError

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(model={self.model})"


def unwrap_observation_data(observation_data: List[ObservationData]) -> TModelPredict:
    """Converts observation data to the format for model prediction outputs.
    That format assumes each observation data has the same set of metrics.
    """
    metrics = set(observation_data[0].metric_names)
    f: TModelMean = {metric: [] for metric in metrics}
    cov: TModelCov = {m1: {m2: [] for m2 in metrics} for m1 in metrics}
    for od in observation_data:
        if set(od.metric_names) != metrics:
            raise ValueError(
                "Each ObservationData should use same set of metrics. "
                "Expected {exp}, got {got}.".format(
                    exp=metrics, got=set(od.metric_names)
                )
            )
        for i, m1 in enumerate(od.metric_names):
            f[m1].append(od.means[i])
            for j, m2 in enumerate(od.metric_names):
                cov[m1][m2].append(od.covariance[i, j])
    return f, cov


def gen_arms(
    observation_features: List[ObservationFeatures],
    arms_by_signature: Optional[Dict[str, Arm]] = None,
) -> Tuple[List[Arm], Optional[Dict[str, TCandidateMetadata]]]:
    """Converts observation features to a tuple of arms list and candidate metadata
    dict, where arm signatures are mapped to their respective candidate metadata.
    """
    # TODO(T34225939): handle static context (which is stored on observation_features)
    arms = []
    candidate_metadata = {}
    for of in observation_features:
        arm = Arm(parameters=of.parameters)
        if arms_by_signature is not None and arm.signature in arms_by_signature:
            existing_arm = arms_by_signature[arm.signature]
            arm = Arm(name=existing_arm.name, parameters=existing_arm.parameters)
        arms.append(arm)
        if of.metadata:
            candidate_metadata[arm.signature] = of.metadata
    return arms, candidate_metadata or None  # None if empty cand. metadata.


def clamp_observation_features(
    observation_features: List[ObservationFeatures], search_space: SearchSpace
) -> List[ObservationFeatures]:
    range_parameters = [
        p for p in search_space.parameters.values() if isinstance(p, RangeParameter)
    ]
    for obsf in observation_features:
        for p in range_parameters:
            if p.name not in obsf.parameters:
                continue
            if p.parameter_type == ParameterType.FLOAT:
                val = checked_cast(float, obsf.parameters[p.name])
            else:
                val = checked_cast(int, obsf.parameters[p.name])
            if val < p.lower:
                logger.info(
                    f"Untransformed parameter {val} "
                    f"less than lower bound {p.lower}, clamping"
                )
                obsf.parameters[p.name] = p.lower
            elif val > p.upper:
                logger.info(
                    f"Untransformed parameter {val} "
                    f"greater than upper bound {p.upper}, clamping"
                )
                obsf.parameters[p.name] = p.upper
    return observation_features


def _get_status_quo_by_trial(
    observations: List[Observation],
    status_quo_name: Optional[str] = None,
    status_quo_features: Optional[ObservationFeatures] = None,
) -> Optional[Dict[int, ObservationData]]:
    r"""
    Given a status quo observation, return a dictionary of trial index to
    the status quo observation data of each trial.

    When either `status_quo_name` or `status_quo_features` exists, return the dict;
    when both exist, use `status_quo_name`;
    when neither exists, return None.

    Args:
        observations: List of observations.
        status_quo_name: Name of the status quo.
        status_quo_features: ObservationFeatures for the status quo.

    Returns:
        A map from trial index to status quo observation data, or None
    """
    trial_idx_to_sq_data = None
    if status_quo_name is not None:
        # identify status quo by arm name
        trial_idx_to_sq_data = {
            int(not_none(obs.features.trial_index)): obs.data
            for obs in observations
            if obs.arm_name == status_quo_name
        }
    elif status_quo_features is not None:
        # identify status quo by (untransformed) feature
        status_quo_signature = json.dumps(
            status_quo_features.parameters, sort_keys=True
        )
        trial_idx_to_sq_data = {
            int(not_none(obs.features.trial_index)): obs.data
            for obs in observations
            if json.dumps(obs.features.parameters, sort_keys=True)
            == status_quo_signature
        }

    return trial_idx_to_sq_data



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from logging import Logger
from typing import Any, Callable, Dict, Optional, Tuple, Type

from ax.core.runner import Runner
from ax.runners.synthetic import SyntheticRunner
from ax.storage.json_store.encoders import runner_to_dict
from ax.storage.json_store.registry import (
    CORE_DECODER_REGISTRY,
    CORE_ENCODER_REGISTRY,
    TDecoderRegistry,
)
from ax.storage.utils import stable_hash
from ax.utils.common.logger import get_logger

logger: Logger = get_logger(__name__)

# """
# Mapping of Runner classes to ints.

# All runners will be stored in the same table in the database. When
# saving, we look up the runner subclass in RUNNER_REGISTRY, and store
# the corresponding type field in the database. When loading, we look
# up the type field in REVERSE_RUNNER_REGISTRY, and initialize the
# corresponding runner subclass.
# """
CORE_RUNNER_REGISTRY: Dict[Type[Runner], int] = {SyntheticRunner: 0}


# pyre-fixme[3]: Return annotation cannot contain `Any`.
def register_runner(
    runner_cls: Type[Runner],
    runner_registry: Dict[Type[Runner], int] = CORE_RUNNER_REGISTRY,
    # pyre-fixme[2]: Parameter annotation cannot contain `Any`.
    # pyre-fixme[24]: Generic type `type` expects 1 type parameter, use
    #  `typing.Type` to avoid runtime subscripting errors.
    encoder_registry: Dict[
        Type, Callable[[Any], Dict[str, Any]]
    ] = CORE_ENCODER_REGISTRY,
    decoder_registry: TDecoderRegistry = CORE_DECODER_REGISTRY,
    val: Optional[int] = None,
    # pyre-fixme[24]: Generic type `type` expects 1 type parameter, use `typing.Type` to
    #  avoid runtime subscripting errors.
) -> Tuple[
    Dict[Type[Runner], int],
    Dict[Type, Callable[[Any], Dict[str, Any]]],
    TDecoderRegistry,
]:
    """Add a custom runner class to the SQA and JSON registries.
    For the SQA registry, if no int is specified, use a hash of the class name.
    """
    registered_val = val or abs(stable_hash(runner_cls.__name__)) % (10**5)

    new_runner_registry = {runner_cls: registered_val, **runner_registry}
    new_encoder_registry = {runner_cls: runner_to_dict, **encoder_registry}
    new_decoder_registry = {runner_cls.__name__: runner_cls, **decoder_registry}

    return new_runner_registry, new_encoder_registry, new_decoder_registry


# pyre-fixme[3]: Return annotation cannot contain `Any`.
def register_runners(
    runner_clss: Dict[Type[Runner], Optional[int]],
    runner_registry: Dict[Type[Runner], int] = CORE_RUNNER_REGISTRY,
    # pyre-fixme[2]: Parameter annotation cannot contain `Any`.
    # pyre-fixme[24]: Generic type `type` expects 1 type parameter, use
    #  `typing.Type` to avoid runtime subscripting errors.
    encoder_registry: Dict[
        Type, Callable[[Any], Dict[str, Any]]
    ] = CORE_ENCODER_REGISTRY,
    decoder_registry: TDecoderRegistry = CORE_DECODER_REGISTRY,
    # pyre-fixme[24]: Generic type `type` expects 1 type parameter, use `typing.Type` to
    #  avoid runtime subscripting errors.
) -> Tuple[
    Dict[Type[Runner], int],
    Dict[Type, Callable[[Any], Dict[str, Any]]],
    TDecoderRegistry,
]:
    """Add custom runner classes to the SQA and JSON registries.
    For the SQA registry, if no int is specified, use a hash of the class name.
    """
    new_runner_registry = {
        **{
            runner_cls: val if val else abs(stable_hash(runner_cls.__name__)) % (10**5)
            for runner_cls, val in runner_clss.items()
        },
        **runner_registry,
    }
    new_encoder_registry = {
        **{runner_cls: runner_to_dict for runner_cls in runner_clss},
        **encoder_registry,
    }
    new_decoder_registry = {
        **{runner_cls.__name__: runner_cls for runner_cls in runner_clss},
        **decoder_registry,
    }

    return new_runner_registry, new_encoder_registry, new_decoder_registry



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


from numbers import Real
from typing import Tuple

# type aliases
TRGB = Tuple[Real, ...]


def rgba(rgb_tuple: TRGB, alpha: float = 1) -> str:
    """Convert RGB tuple to an RGBA string."""
    return "rgba({},{},{},{alpha})".format(*rgb_tuple, alpha=alpha)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import json
from typing import Union

import torch
from ax.utils.common.typeutils import checked_cast


def torch_type_to_str(value: Union[torch.dtype, torch.device, torch.Size]) -> str:
    """Converts torch types, commonly used in Ax, to string representations."""
    if isinstance(value, torch.dtype):
        return str(value)
    if isinstance(value, torch.device):
        return checked_cast(str, value.type)
    if isinstance(value, torch.Size):
        return json.dumps(list(value))
    raise ValueError(f"Object {value} was of unexpected torch type.")


def torch_type_from_str(
    identifier: str, type_name: str
) -> Union[torch.dtype, torch.device, torch.Size]:
    if type_name == "device":
        return torch.device(identifier)
    if type_name == "dtype":
        return getattr(torch, identifier[6:])
    if type_name == "Size":
        return torch.Size(json.loads(identifier))
    raise ValueError(f"Unexpected type: {type_name} for identifier: {identifier}.")



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import numpy as np
from ax.core.observation import ObservationData
from ax.modelbridge.transforms.ivw import IVW, ivw_metric_merge
from ax.utils.common.testutils import TestCase


class IVWTransformTest(TestCase):
    def test_NoRepeats(self) -> None:
        obsd = ObservationData(
            metric_names=["m1", "m2"],
            means=np.array([1.0, 2.0]),
            covariance=np.array([[1.0, 0.2], [0.2, 2.0]]),
        )
        obsd2 = ivw_metric_merge(obsd)
        self.assertEqual(obsd2, obsd)

    def test_Merge(self) -> None:
        obsd = ObservationData(
            metric_names=["m1", "m2", "m2"],
            means=np.array([1.0, 2.0, 1.0]),
            covariance=np.array([[1.0, 0.2, 0.4], [0.2, 2.0, 0.8], [0.4, 0.8, 3.0]]),
        )
        obsd2 = ivw_metric_merge(obsd)
        self.assertEqual(obsd2.metric_names, ["m1", "m2"])
        self.assertTrue(np.array_equal(obsd2.means, np.array([1.0, 0.6 * 2 + 0.4])))
        cov12 = 0.2 * 0.6 + 0.4 * 0.4
        # var(w1*y1 + w2*y2) =
        # w1 ** 2 * var(y1) + w2 ** 2 * var(y2) + 2 * w1 * w2 * cov(y1, y2)
        cov22 = 0.6**2 * 2.0 + 0.4**2 * 3 + 2 * 0.6 * 0.4 * 0.8
        cov_true = np.array([[1.0, cov12], [cov12, cov22]])
        discrep = np.max(np.abs(obsd2.covariance - cov_true))
        self.assertTrue(discrep < 1e-8)

    def test_NoiselessMerge(self) -> None:
        # One noiseless
        obsd = ObservationData(
            metric_names=["m1", "m2", "m2"],
            means=np.array([1.0, 2.0, 1.0]),
            covariance=np.array([[1.0, 0.2, 0.4], [0.2, 2.0, 0.8], [0.4, 0.8, 0.0]]),
        )
        obsd2 = ivw_metric_merge(obsd)
        np.array_equal(obsd2.means, np.array([1.0, 1.0]))
        cov_true = np.array([[1.0, 0.4], [0.4, 0.0]])
        self.assertTrue(np.array_equal(obsd2.covariance, cov_true))
        # Conflicting noiseless, default (warn)
        obsd = ObservationData(
            metric_names=["m1", "m2", "m2"],
            means=np.array([1.0, 2.0, 1.0]),
            covariance=np.array([[1.0, 0.2, 0.4], [0.2, 0.0, 0.8], [0.4, 0.8, 0.0]]),
        )
        with self.assertRaises(ValueError):
            obsd2 = ivw_metric_merge(obsd, conflicting_noiseless="wrong")
        obsd2 = ivw_metric_merge(obsd)
        self.assertTrue(np.array_equal(obsd2.means, np.array([1.0, 2.0])))
        cov_true = np.array([[1.0, 0.2], [0.2, 0.0]])
        self.assertTrue(np.array_equal(obsd2.covariance, cov_true))
        # Conflicting noiseless, raise
        with self.assertRaises(ValueError):
            obsd2 = ivw_metric_merge(obsd, conflicting_noiseless="raise")

    def test_Transform(self) -> None:
        obsd1_0 = ObservationData(
            metric_names=["m1", "m2", "m2"],
            means=np.array([1.0, 2.0, 1.0]),
            covariance=np.array([[1.0, 0.2, 0.4], [0.2, 2.0, 0.8], [0.4, 0.8, 3.0]]),
        )
        obsd1_1 = ObservationData(
            metric_names=["m1", "m1", "m2", "m2"],
            means=np.array([1.0, 1.0, 2.0, 1.0]),
            covariance=np.array(
                [
                    [1.0, 0.0, 0.0, 0.0],
                    [0.0, 1.0, 0.2, 0.4],
                    [0.0, 0.2, 2.0, 0.8],
                    [0.0, 0.4, 0.8, 3.0],
                ]
            ),
        )
        obsd2_0 = ObservationData(
            metric_names=["m1", "m2"],
            means=np.array([1.0, 1.6]),
            covariance=np.array([[1.0, 0.28], [0.28, 1.584]]),
        )
        obsd2_1 = ObservationData(
            metric_names=["m1", "m2"],
            means=np.array([1.0, 1.6]),
            covariance=np.array([[0.5, 0.14], [0.14, 1.584]]),
        )
        observation_data = [obsd1_0, obsd1_1]
        t = IVW(None, [])
        observation_data2 = t._transform_observation_data(observation_data)
        observation_data2_true = [obsd2_0, obsd2_1]
        for i, obsd in enumerate(observation_data2_true):
            self.assertEqual(observation_data2[i].metric_names, obsd.metric_names)
            self.assertTrue(np.array_equal(observation_data2[i].means, obsd.means))
            discrep = np.max(np.abs(observation_data2[i].covariance - obsd.covariance))
            self.assertTrue(discrep < 1e-8)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from datetime import datetime, timedelta
from time import time
from typing import Generator

import pandas as pd


DS_FRMT = "%Y-%m-%d"  # Format to use for parsing DS strings.


def to_ds(ts: datetime) -> str:
    """Convert a `datetime` to a DS string."""
    return datetime.strftime(ts, DS_FRMT)


def to_ts(ds: str) -> datetime:
    """Convert a DS string to a `datetime`."""
    return datetime.strptime(ds, DS_FRMT)


def _ts_to_pandas(ts: int) -> pd.Timestamp:
    """Convert int timestamp into pandas timestamp."""
    return pd.Timestamp(datetime.fromtimestamp(ts))


def _pandas_ts_to_int(ts: pd.Timestamp) -> int:
    """Convert int timestamp into pandas timestamp."""
    # pyre-fixme[7]: Expected `int` but got `float`.
    return ts.to_pydatetime().timestamp()


def current_timestamp_in_millis() -> int:
    """Grab current timestamp in milliseconds as an int."""
    return int(round(time() * 1000))


def timestamps_in_range(
    start: datetime, end: datetime, delta: timedelta
) -> Generator[datetime, None, None]:
    """Generator of timestamps in range [start, end], at intervals
    delta.
    """
    curr = start
    while curr <= end:
        yield curr
        curr += delta


def unixtime_to_pandas_ts(ts: float) -> pd.Timestamp:
    """Convert float unixtime into pandas timestamp (UTC)."""
    return pd.to_datetime(ts, unit="s")



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import itertools
from copy import deepcopy
from unittest import mock
from unittest.mock import Mock, patch

import pandas as pd
from ax.core.base_trial import BaseTrial, TrialStatus
from ax.core.data import Data
from ax.core.generator_run import GeneratorRun, GeneratorRunType
from ax.core.runner import Runner
from ax.exceptions.core import UnsupportedError, UserInputError
from ax.runners.synthetic import SyntheticRunner
from ax.utils.common.result import Ok
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import (
    get_arms,
    get_branin_arms,
    get_experiment,
    get_objective,
    get_test_map_data_experiment,
)

TEST_DATA = Data(
    df=pd.DataFrame(
        [
            {
                "arm_name": "0_0",
                "metric_name": get_objective().metric.name,
                "mean": 1.0,
                "sem": 2.0,
                "trial_index": 0,
            }
        ]
    )
)


class TrialTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.mock_supports_trial_type = mock.patch(
            f"{get_experiment.__module__}.Experiment.supports_trial_type",
            return_value=True,
        )
        self.mock_supports_trial_type.start()
        self.experiment = get_experiment()
        self.trial = self.experiment.new_trial(ttl_seconds=123, trial_type="foo")
        self.trial.update_run_metadata(metadata={"foo": "bar"})
        self.trial.update_stop_metadata(metadata={"bar": "baz"})
        self.arm = get_arms()[0]
        self.trial.add_arm(self.arm)

    def tearDown(self) -> None:
        self.mock_supports_trial_type.stop()

    def test__validate_can_attach_data(self) -> None:
        self.trial.mark_running(no_runner_required=True)
        self.trial.mark_completed()

        expected_msg = (
            "Trial 0 has already been completed with data. To add more data to "
        )
        with self.assertRaisesRegex(UnsupportedError, expected_msg):
            self.trial._validate_can_attach_data()

    def test_eq(self) -> None:
        new_trial = self.experiment.new_trial()
        self.assertNotEqual(self.trial, new_trial)

    def test_basic_properties(self) -> None:
        self.assertEqual(self.experiment, self.trial.experiment)
        self.assertEqual(self.trial.index, 0)
        self.assertEqual(self.trial.status, TrialStatus.CANDIDATE)
        self.assertTrue(self.trial.status.is_candidate)
        self.assertIsNotNone(self.trial.time_created)
        self.assertEqual(self.trial.arms_by_name["0_0"], self.trial.arm)
        self.assertEqual(self.trial.arms[0].signature, self.arm.signature)
        self.assertEqual(self.trial.abandoned_arms, [])
        self.assertEqual(
            self.trial.generator_run.generator_run_type, GeneratorRunType.MANUAL.name
        )

        # Test empty arms
        with self.assertRaises(AttributeError):
            self.experiment.new_trial().arm_weights

        self.trial.mark_running(no_runner_required=True)
        self.assertTrue(self.trial.status.is_running)

        self.trial._status = TrialStatus.COMPLETED
        self.assertEqual(str(self.trial._status), "TrialStatus.COMPLETED")
        self.assertTrue(self.trial.status.is_completed)
        self.assertTrue(self.trial.completed_successfully)

    def test_adding_new_trials(self) -> None:
        new_arm = get_arms()[1]
        cand_metadata = {new_arm.signature: {"a": "b"}}
        new_trial = self.experiment.new_trial(
            generator_run=GeneratorRun(
                arms=[new_arm],
                # pyre-fixme[6]: For 2nd param expected `Optional[Dict[str,
                #  Optional[Dict[str, typing.Any]]]]` but got `Dict[str, Dict[str,
                #  str]]`.
                candidate_metadata_by_arm_signature=cand_metadata,
            )
        )
        with self.assertRaises(ValueError):
            self.experiment.new_trial(generator_run=GeneratorRun(arms=get_arms()))
        self.assertEqual(new_trial.arms_by_name["1_0"].signature, new_arm.signature)
        with self.assertRaises(KeyError):
            self.trial.arms_by_name["1_0"]
        self.assertEqual(
            new_trial._get_candidate_metadata_from_all_generator_runs(),
            {"1_0": cand_metadata[new_arm.signature]},
        )
        self.assertEqual(
            new_trial._get_candidate_metadata("1_0"), cand_metadata[new_arm.signature]
        )
        self.assertRaises(
            ValueError, new_trial._get_candidate_metadata, "this_is_not_an_arm"
        )

    def test_add_trial_same_arm(self) -> None:
        # Check that adding new arm w/out name works correctly.
        new_trial1 = self.experiment.new_trial(
            generator_run=GeneratorRun(arms=[self.arm.clone(clear_name=True)])
        )
        self.assertEqual(new_trial1.arm.name, self.trial.arm.name)
        self.assertFalse(new_trial1.arm is self.trial.arm)
        # Check that adding new arm with name works correctly.
        new_trial2 = self.experiment.new_trial(
            generator_run=GeneratorRun(arms=[self.arm.clone()])
        )
        self.assertEqual(new_trial2.arm.name, self.trial.arm.name)
        self.assertFalse(new_trial2.arm is self.trial.arm)
        arm_wrong_name = self.arm.clone(clear_name=True)
        arm_wrong_name.name = "wrong_name"
        with self.assertRaises(ValueError):
            new_trial2 = self.experiment.new_trial(
                generator_run=GeneratorRun(arms=[arm_wrong_name])
            )

    def test_abandonment(self) -> None:
        self.assertFalse(self.trial.status.is_abandoned)
        self.trial.mark_abandoned(reason="testing")
        self.assertTrue(self.trial.status.is_abandoned)
        self.assertFalse(self.trial.status.is_failed)
        self.assertTrue(self.trial.did_not_complete)

    def test_failed(self) -> None:
        fail_reason = "testing"
        self.trial.runner = SyntheticRunner()
        self.trial.run()
        self.trial.mark_failed(reason=fail_reason)

        self.assertTrue(self.trial.status.is_failed)
        self.assertTrue(self.trial.did_not_complete)
        self.assertEqual(self.trial.failed_reason, fail_reason)

    def test_trial_run_does_not_overwrite_existing_metadata(self) -> None:
        self.trial.runner = SyntheticRunner(dummy_metadata="y")
        self.trial.update_run_metadata({"orig_metadata": "x"})
        self.trial.run()
        self.assertDictEqual(
            self.trial.run_metadata,
            {
                "name": "test_0",
                "orig_metadata": "x",
                "dummy_metadata": "y",
                # this is set in setUp
                "foo": "bar",
            },
        )

    def test_mark_as(self) -> None:
        for terminal_status in (
            TrialStatus.ABANDONED,
            TrialStatus.FAILED,
            TrialStatus.COMPLETED,
            TrialStatus.EARLY_STOPPED,
        ):
            self.setUp()
            # Note: This only tests the no-runner case (and thus not staging)
            for status in (TrialStatus.RUNNING, terminal_status):
                kwargs = {}
                if status == TrialStatus.RUNNING:
                    kwargs["no_runner_required"] = True
                if status == TrialStatus.ABANDONED:
                    kwargs["reason"] = "test_reason_abandon"
                if status == TrialStatus.FAILED:
                    kwargs["reason"] = "test_reason_failed"
                self.trial.mark_as(status=status, **kwargs)
                self.assertTrue(self.trial.status == status)

                if status == TrialStatus.ABANDONED:
                    self.assertEqual(self.trial.abandoned_reason, "test_reason_abandon")
                    self.assertIsNone(self.trial.failed_reason)
                elif status == TrialStatus.FAILED:
                    self.assertEqual(self.trial.failed_reason, "test_reason_failed")
                    self.assertIsNone(self.trial.abandoned_reason)

                if status != TrialStatus.RUNNING:
                    self.assertTrue(self.trial.status.is_terminal)

                if status in [
                    TrialStatus.RUNNING,
                    TrialStatus.EARLY_STOPPED,
                    TrialStatus.COMPLETED,
                ]:
                    self.assertTrue(self.trial.status.expecting_data)
                else:
                    self.assertFalse(self.trial.status.expecting_data)

    def test_stop(self) -> None:
        # test bad old status
        with self.assertRaisesRegex(ValueError, "Can only stop STAGED or RUNNING"):
            self.trial.stop(new_status=TrialStatus.ABANDONED)

        # test bad new status
        self.trial.mark_running(no_runner_required=True)
        with self.assertRaisesRegex(ValueError, "New status of a stopped trial must"):
            self.trial.stop(new_status=TrialStatus.CANDIDATE)

        # dummy runner for testing stopping functionality
        class DummyStopRunner(Runner):
            def run(self, trial):
                pass

            def stop(self, trial, reason):
                return {"reason": reason} if reason else {}

        # test valid stopping
        for reason, new_status in itertools.product(
            (None, "because"),
            (TrialStatus.COMPLETED, TrialStatus.ABANDONED, TrialStatus.EARLY_STOPPED),
        ):
            self.setUp()
            self.trial._runner = DummyStopRunner()
            self.trial.mark_running()
            self.assertEqual(self.trial.status, TrialStatus.RUNNING)
            self.trial.stop(new_status=new_status, reason=reason)
            self.assertEqual(self.trial.status, new_status)
            self.assertEqual(
                self.trial.stop_metadata, {} if reason is None else {"reason": reason}
            )

    @patch(
        f"{BaseTrial.__module__}.{BaseTrial.__name__}.lookup_data",
        return_value=TEST_DATA,
    )
    # pyre-fixme[3]: Return type must be annotated.
    def test_objective_mean(self, _mock):
        self.assertEqual(self.trial.objective_mean, 1.0)

    @patch(
        f"{BaseTrial.__module__}.{BaseTrial.__name__}.lookup_data", return_value=Data()
    )
    # pyre-fixme[3]: Return type must be annotated.
    def test_objective_mean_empty_df(self, _mock):
        with self.assertRaisesRegex(ValueError, "not yet in data for trial."):
            self.assertIsNone(self.trial.objective_mean)

    @patch(
        f"{BaseTrial.__module__}.{BaseTrial.__name__}.fetch_data_results",
        return_value={get_objective().metric.name: Ok(TEST_DATA)},
    )
    def test_fetch_data_result(self, mock: Mock) -> None:
        metric_name = get_objective().metric.name
        results = self.experiment.trials[0].fetch_data_results()

        self.assertTrue(results[metric_name].is_ok())
        self.assertEqual(
            results[metric_name].ok, self.experiment.trials[0].fetch_data()
        )

    def test_Repr(self) -> None:
        repr_ = (
            "Trial(experiment_name='test', index=0, "
            "status=TrialStatus.CANDIDATE, arm=Arm(name='0_0', "
            "parameters={'w': 0.85, 'x': 1, 'y': 'baz', 'z': False}))"
        )
        self.assertEqual(str(self.trial), repr_)

    def test_update_run_metadata(self) -> None:
        self.assertEqual(len(self.trial.run_metadata), 1)
        old_run_metadata = deepcopy(self.trial.run_metadata)
        self.trial.update_run_metadata({"something": "new"})
        self.assertDictEqual(
            self.trial.run_metadata, {**old_run_metadata, "something": "new"}
        )

    def test_update_stop_metadata(self) -> None:
        self.assertEqual(len(self.trial.stop_metadata), 1)
        old_stop_metadata = deepcopy(self.trial.stop_metadata)
        self.trial.update_stop_metadata({"something": "new"})
        self.assertEqual(
            self.trial.stop_metadata, {**old_stop_metadata, "something": "new"}
        )

    def test_update_trial_data(self) -> None:
        # Verify components before we attach trial data
        self.assertEqual(1, len(self.trial.arms))
        arm_name = self.trial.arm.name

        self.assertEqual(
            2,
            len(self.trial.experiment.metrics)
            - len(self.trial.experiment.tracking_metrics),
        )
        self.assertTrue("m1" in self.trial.experiment.metrics)
        self.assertTrue("m2" in self.trial.experiment.metrics)

        data = self.trial.lookup_data().df.to_dict(orient="index")
        self.assertTrue(len(data) == 0)

        # Attach data
        self.trial.update_trial_data(raw_data={"m1": 1.0, "m2": 2.0})

        # Confirm the expected state after attaching data
        data = (
            self.trial.lookup_data()
            .df.set_index(["arm_name", "metric_name"])
            .to_dict(orient="index")
        )

        self.assertEqual(1.0, data[(arm_name, "m1")]["mean"])
        self.assertEqual(2.0, data[(arm_name, "m2")]["mean"])

        # Try to attach MapData.
        with self.assertRaisesRegex(
            UserInputError,
            "The format of the `raw_data` is not compatible with `Data`. ",
        ):
            self.trial.update_trial_data(raw_data=[({"time": 0}, {"m1": 1.0})])

        # Try to attach Data to a MapData experiment.
        map_experiment = get_test_map_data_experiment(
            num_trials=3, num_fetches=2, num_complete=1
        )
        map_trial = map_experiment.new_trial().add_arm(
            arm=get_branin_arms(n=1, seed=0)[0]
        )
        map_trial.update_trial_data(raw_data=[({"time": 0}, {"m1": 1.0})])
        with self.assertRaisesRegex(
            UserInputError,
            "The format of the `raw_data` is not compatible with `MapData`. ",
        ):
            map_trial.update_trial_data(raw_data={"m1": 1.0})

        # Error if the MapData inputs are not formatted correctly.
        with self.assertRaisesRegex(
            UserInputError,
            "Raw data does not conform to the expected structure.",
        ):
            map_trial.update_trial_data(raw_data=[["aa", {"m1": 1.0}]])

    def test_clone_to(self) -> None:
        # cloned trial attached to the same experiment
        self.trial.mark_running(no_runner_required=True)
        new_trial = self.trial.clone_to()
        self.assertIs(new_trial.experiment, self.trial.experiment)
        # Test equality of all attributes except index, time_created, and experiment.
        for k, v in new_trial.__dict__.items():
            if k in ["_index", "_time_created", "_experiment", "_time_run_started"]:
                continue
            self.assertEqual(v, self.trial.__dict__[k])

        # cloned trial attached to a new experiment
        new_experiment = get_experiment()
        new_trial = self.trial.clone_to(new_experiment)
        self.assertEqual(new_trial, self.trial)

        # make sure updating cloned trial doesn't affect original one
        new_trial._status = TrialStatus.COMPLETED
        self.assertTrue(new_trial.status.is_completed)
        self.assertFalse(self.trial.status.is_completed)

    def test_update_trial_status_on_clone(self) -> None:
        for status in [
            TrialStatus.CANDIDATE,
            TrialStatus.STAGED,
            TrialStatus.RUNNING,
            TrialStatus.EARLY_STOPPED,
            TrialStatus.COMPLETED,
            TrialStatus.FAILED,
            TrialStatus.ABANDONED,
        ]:
            self.trial._failed_reason = self.trial._abandoned_reason = None
            if status != TrialStatus.CANDIDATE:
                self.trial.mark_as(
                    status=status, unsafe=True, no_runner_required=True, reason="test"
                )
            test_trial = self.trial.clone_to()
            # Overwrite unimportant attrs before equality check.
            test_trial._index = self.trial.index
            test_trial._time_created = self.trial._time_created
            test_trial._time_staged = self.trial._time_staged
            self.assertEqual(self.trial, test_trial)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from typing import Set, Tuple
from unittest.mock import MagicMock, Mock, patch

import torch
from ax.exceptions.core import UnsupportedError
from ax.models.torch.utils import (
    _get_X_pending_and_observed,
    get_botorch_objective_and_transform,
)
from ax.utils.common.testutils import TestCase
from ax.utils.common.typeutils import checked_cast, not_none
from botorch.acquisition.knowledge_gradient import qKnowledgeGradient
from botorch.acquisition.logei import qLogNoisyExpectedImprovement
from botorch.acquisition.monte_carlo import qNoisyExpectedImprovement
from botorch.acquisition.multi_objective.monte_carlo import (
    qNoisyExpectedHypervolumeImprovement,
)
from botorch.acquisition.multi_objective.multi_output_risk_measures import (
    MARS,
    MultiOutputExpectation,
)
from botorch.acquisition.multi_objective.objective import WeightedMCMultiOutputObjective
from botorch.acquisition.objective import (
    ConstrainedMCObjective,
    GenericMCObjective,
    ScalarizedPosteriorTransform,
)
from botorch.acquisition.risk_measures import Expectation
from botorch.exceptions.errors import BotorchTensorDimensionError
from botorch.models.model import Model


class TorchUtilsTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.device = torch.device("cpu")
        self.dtype = torch.double
        self.mock_botorch_model = MagicMock(Model)
        tkwargs = {"dtype": self.dtype, "device": self.device}
        self.X_dummy = torch.ones(1, 3, **tkwargs)
        self.outcome_constraints = (
            torch.tensor([[1.0]], **tkwargs),
            torch.tensor([[0.5]], **tkwargs),
        )
        self.objective_weights = torch.ones(1, **tkwargs)
        self.moo_objective_weights = torch.ones(2, **tkwargs)
        self.objective_thresholds = torch.tensor([0.5, 1.5], **tkwargs)

    def test_get_X_pending_and_observed(self) -> None:
        def _to_obs_set(X: torch.Tensor) -> Set[Tuple[float]]:
            return {tuple(float(x_i) for x_i in x) for x in X}

        # Apply filter normally
        Xs = [torch.tensor([[0.0, 0.0], [0.0, 1.0]])]
        objective_weights = torch.tensor([1.0])
        bounds = [(0.0, 1.0), (0.0, 1.0)]
        fixed_features = {1: 1.0}
        _, X_observed = _get_X_pending_and_observed(
            Xs=Xs,
            objective_weights=self.objective_weights,
            bounds=bounds,
            fixed_features=fixed_features,
        )
        expected = Xs[0][1:]
        self.assertEqual(_to_obs_set(expected), _to_obs_set(not_none(X_observed)))

        # Filter too strict; return unfiltered X_observed
        fixed_features = {0: 1.0}
        _, X_observed = _get_X_pending_and_observed(
            Xs=Xs,
            objective_weights=objective_weights,
            bounds=bounds,
            fixed_features=fixed_features,
        )
        expected = Xs[0]
        self.assertEqual(_to_obs_set(expected), _to_obs_set(not_none(X_observed)))

        # Out of design observations are filtered out
        Xs = [torch.tensor([[2.0, 3.0], [3.0, 4.0]])]
        _, X_observed = _get_X_pending_and_observed(
            Xs=Xs,
            objective_weights=objective_weights,
            bounds=bounds,
            fixed_features=fixed_features,
            fit_out_of_design=False,
        )
        self.assertIsNone(X_observed)

        # Keep out of design observations
        _, X_observed = _get_X_pending_and_observed(
            Xs=Xs,
            objective_weights=objective_weights,
            bounds=bounds,
            fixed_features=fixed_features,
            fit_out_of_design=True,
        )
        expected = Xs[0]
        self.assertEqual(_to_obs_set(expected), _to_obs_set(not_none(X_observed)))

    @patch(
        f"{get_botorch_objective_and_transform.__module__}.get_infeasible_cost",
        return_value=1.0,
    )
    def test_get_botorch_objective(self, _) -> None:
        # For KG with outcome constraints, use a ConstrainedMCObjective
        obj, tf = get_botorch_objective_and_transform(
            botorch_acqf_class=qKnowledgeGradient,
            model=self.mock_botorch_model,
            outcome_constraints=self.outcome_constraints,
            objective_weights=self.objective_weights,
            X_observed=self.X_dummy,
        )
        self.assertIsInstance(obj, ConstrainedMCObjective)
        self.assertIsNone(tf)

        # SampleReducingMCAcquisitionFunctions with outcome constraints, handle
        # the constraints separately
        obj, tf = get_botorch_objective_and_transform(
            botorch_acqf_class=qLogNoisyExpectedImprovement,
            model=self.mock_botorch_model,
            outcome_constraints=self.outcome_constraints,
            objective_weights=self.objective_weights,
            X_observed=self.X_dummy,
        )
        self.assertIsInstance(obj, GenericMCObjective)
        self.assertIsNone(tf)

        # By default, `ScalarizedPosteriorTransform` should be picked in absence of
        # outcome constraints.
        obj, tf = get_botorch_objective_and_transform(
            botorch_acqf_class=qNoisyExpectedImprovement,
            model=self.mock_botorch_model,
            objective_weights=self.objective_weights,
            X_observed=self.X_dummy,
        )
        self.assertIsNone(obj)
        self.assertIsInstance(tf, ScalarizedPosteriorTransform)

        # Test MOO case.
        with self.assertRaises(BotorchTensorDimensionError):
            get_botorch_objective_and_transform(
                botorch_acqf_class=qNoisyExpectedHypervolumeImprovement,
                model=self.mock_botorch_model,
                objective_weights=self.objective_weights,  # Only has 1 objective.
                X_observed=self.X_dummy,
            )

        obj, tf = get_botorch_objective_and_transform(
            botorch_acqf_class=qNoisyExpectedHypervolumeImprovement,
            model=self.mock_botorch_model,
            objective_weights=self.moo_objective_weights,  # Has 2 objectives.
            X_observed=self.X_dummy,
        )
        self.assertIsInstance(obj, WeightedMCMultiOutputObjective)
        self.assertIsNone(tf)

    @patch(f"{MARS.__module__}.MARS.set_baseline_Y")
    def test_get_botorch_objective_w_risk_measures(
        self, mock_set_baseline_Y: Mock
    ) -> None:
        # Test outcome constraint error.
        with self.assertRaisesRegex(NotImplementedError, "Outcome constraints"):
            get_botorch_objective_and_transform(
                botorch_acqf_class=qNoisyExpectedImprovement,
                model=self.mock_botorch_model,
                objective_weights=self.objective_weights,
                outcome_constraints=self.outcome_constraints,
                risk_measure=Expectation(n_w=5),
            )
        # Test user supplied preprocessing function error.
        with self.assertRaisesRegex(UnsupportedError, "User supplied"):
            get_botorch_objective_and_transform(
                botorch_acqf_class=qNoisyExpectedImprovement,
                model=self.mock_botorch_model,
                objective_weights=self.objective_weights,
                risk_measure=Expectation(
                    n_w=5,
                    preprocessing_function=WeightedMCMultiOutputObjective(
                        weights=torch.tensor([1.0, 1.0])
                    ),
                ),
            )
        # Test MARS errors.
        with self.assertRaisesRegex(UnsupportedError, "X_observed"):
            get_botorch_objective_and_transform(
                botorch_acqf_class=qNoisyExpectedImprovement,
                model=self.mock_botorch_model,
                objective_weights=self.moo_objective_weights,
                risk_measure=MARS(alpha=0.8, n_w=5, chebyshev_weights=[]),
            )
        # Test single objective case.
        risk_measure, _ = get_botorch_objective_and_transform(
            botorch_acqf_class=qNoisyExpectedImprovement,
            model=self.mock_botorch_model,
            objective_weights=torch.tensor([-1.0]),
            risk_measure=Expectation(n_w=2),
        )
        self.assertTrue(
            torch.allclose(
                not_none(risk_measure)(torch.tensor([[1.0], [2.0]])),
                torch.tensor([-1.5]),
            )
        )
        # Test scalarized objective with single objective risk measure.
        risk_measure, _ = get_botorch_objective_and_transform(
            botorch_acqf_class=qNoisyExpectedImprovement,
            model=self.mock_botorch_model,
            objective_weights=torch.tensor([-1.0, 1.0, 0.0]),
            risk_measure=Expectation(n_w=2),
        )
        Y = torch.tensor([[1.0, -1.0, 3.0], [2.0, -2.0, 3.0]])
        self.assertTrue(
            torch.allclose(
                not_none(risk_measure)(Y),
                torch.tensor([-3.0]),
            )
        )
        # Test MO risk measure.
        risk_measure, _ = get_botorch_objective_and_transform(
            botorch_acqf_class=qNoisyExpectedHypervolumeImprovement,
            model=self.mock_botorch_model,
            objective_weights=torch.tensor([-1.0, 2.0, 0.0]),
            risk_measure=MultiOutputExpectation(n_w=2),
        )
        self.assertTrue(
            torch.allclose(
                not_none(risk_measure)(Y),
                torch.tensor([-1.5, -3.0]),
            )
        )
        # Test MARS.
        risk_measure, _ = get_botorch_objective_and_transform(
            botorch_acqf_class=qNoisyExpectedImprovement,
            model=self.mock_botorch_model,
            objective_weights=torch.tensor([-1.0, 2.0, 0.0]),
            risk_measure=MARS(
                alpha=0.8,
                n_w=2,
                chebyshev_weights=[],
                baseline_Y=torch.tensor([[0.0, 1.0], [1.0, 0.0]]),
            ),
            X_observed=torch.ones(2, 2),  # dummy
        )
        mock_set_baseline_Y.assert_called_once()
        risk_measure = checked_cast(MARS, risk_measure)
        self.assertIsInstance(
            risk_measure.preprocessing_function, WeightedMCMultiOutputObjective
        )
        ch_weights = risk_measure.chebyshev_weights
        self.assertEqual(ch_weights.shape, torch.Size([2]))
        self.assertTrue(torch.allclose(ch_weights.sum(), torch.tensor(1.0)))
        # Overwrite ch weights to simplify the test.
        risk_measure.chebyshev_weights = [0.0, 1.0]
        self.assertTrue(
            torch.allclose(
                not_none(risk_measure)(Y),
                torch.tensor([-4.0]),
            )
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from ax.models.base import Model
from ax.utils.common.testutils import TestCase


class BaseModelTest(TestCase):
    def test_base_model(self) -> None:
        model = Model()
        raw_state = {"foo": "bar", "two": 3.0}
        self.assertEqual(model.serialize_state(raw_state), raw_state)
        self.assertEqual(model.deserialize_state(raw_state), raw_state)
        self.assertEqual(model._get_state(), {})
        with self.assertRaisesRegex(
            NotImplementedError, "Feature importance not available"
        ):
            model.feature_importances()



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from ax.core.parameter import (
    ChoiceParameter,
    FixedParameter,
    ParameterType,
    RangeParameter,
)
from ax.core.parameter_constraint import (
    ComparisonOp,
    OrderConstraint,
    ParameterConstraint,
    SumConstraint,
)
from ax.utils.common.testutils import TestCase


class ParameterConstraintTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.constraint = ParameterConstraint(
            constraint_dict={"x": 2.0, "y": -3.0}, bound=6.0
        )
        self.constraint_repr = "ParameterConstraint(2.0*x + -3.0*y <= 6.0)"

    def test_Eq(self) -> None:
        constraint1 = ParameterConstraint(
            constraint_dict={"x": 2.0, "y": -3.0}, bound=6.0
        )
        constraint2 = ParameterConstraint(
            constraint_dict={"y": -3.0, "x": 2.0}, bound=6.0
        )
        self.assertEqual(constraint1, constraint2)

        constraint3 = ParameterConstraint(
            constraint_dict={"x": 2.0, "y": -5.0}, bound=6.0
        )
        self.assertNotEqual(constraint1, constraint3)

    def test_Properties(self) -> None:
        self.assertEqual(self.constraint.constraint_dict["x"], 2.0)
        self.assertEqual(self.constraint.bound, 6.0)

    def test_Repr(self) -> None:
        self.assertEqual(str(self.constraint), self.constraint_repr)

    def test_Validate(self) -> None:
        parameters = {"x": 4, "z": 3}
        with self.assertRaises(ValueError):
            # pyre-fixme[6]: For 1st param expected `Dict[str, Union[float, int]]`
            #  but got `Dict[str, int]`.
            self.constraint.check(parameters)

        # check slack constraint
        parameters = {"x": 4, "y": 1}
        # pyre-fixme[6]: For 1st param expected `Dict[str, Union[float, int]]` but
        #  got `Dict[str, int]`.
        self.assertTrue(self.constraint.check(parameters))

        # check tight constraint (within numerical tolerance)
        parameters = {"x": 4, "y": (2 - 0.5e-8) / 3}
        self.assertTrue(self.constraint.check(parameters))

        # check violated constraint
        parameters = {"x": 4, "y": (2 - 0.5e-6) / 3}
        self.assertFalse(self.constraint.check(parameters))

    def test_Clone(self) -> None:
        constraint_clone = self.constraint.clone()
        self.assertEqual(self.constraint.bound, constraint_clone.bound)

        constraint_clone._bound = 7.0
        self.assertNotEqual(self.constraint.bound, constraint_clone.bound)

    def test_CloneWithTransformedParameters(self) -> None:
        constraint_clone = self.constraint.clone_with_transformed_parameters(
            transformed_parameters={}
        )
        self.assertEqual(self.constraint.bound, constraint_clone.bound)

        constraint_clone._bound = 7.0
        self.assertNotEqual(self.constraint.bound, constraint_clone.bound)

    def test_Sortable(self) -> None:
        constraint1 = ParameterConstraint(
            constraint_dict={"x": 2.0, "y": -3.0}, bound=1.0
        )
        constraint2 = ParameterConstraint(
            constraint_dict={"y": -3.0, "x": 2.0}, bound=6.0
        )
        self.assertTrue(constraint1 < constraint2)


class OrderConstraintTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.x = RangeParameter("x", ParameterType.INT, lower=0, upper=1)
        self.y = RangeParameter("y", ParameterType.INT, lower=0, upper=1)
        self.constraint = OrderConstraint(
            lower_parameter=self.x, upper_parameter=self.y
        )
        self.constraint_repr = "OrderConstraint(x <= y)"

    def test_Properties(self) -> None:
        self.assertEqual(self.constraint.lower_parameter.name, "x")
        self.assertEqual(self.constraint.upper_parameter.name, "y")

    def test_Repr(self) -> None:
        self.assertEqual(str(self.constraint), self.constraint_repr)

    def test_Validate(self) -> None:
        self.assertTrue(self.constraint.check({"x": 0, "y": 1}))
        self.assertTrue(self.constraint.check({"x": 1, "y": 1}))
        self.assertFalse(self.constraint.check({"x": 1, "y": 0}))

    def test_Clone(self) -> None:
        constraint_clone = self.constraint.clone()
        self.assertEqual(
            self.constraint.lower_parameter, constraint_clone.lower_parameter
        )

        constraint_clone._lower_parameter = self.y
        self.assertNotEqual(
            self.constraint.lower_parameter, constraint_clone.lower_parameter
        )

    def test_CloneWithTransformedParameters(self) -> None:
        constraint_clone = self.constraint.clone_with_transformed_parameters(
            transformed_parameters={p.name: p for p in self.constraint.parameters}
        )
        self.assertEqual(
            self.constraint.lower_parameter, constraint_clone.lower_parameter
        )

        constraint_clone._lower_parameter = self.y
        self.assertNotEqual(
            self.constraint.lower_parameter, constraint_clone.lower_parameter
        )

    def test_InvalidSetup(self) -> None:
        z = FixedParameter("z", ParameterType.INT, 0)
        with self.assertRaises(ValueError):
            self.constraint = OrderConstraint(lower_parameter=self.x, upper_parameter=z)

        z = ChoiceParameter("z", ParameterType.STRING, ["a", "b", "c"])
        with self.assertRaises(ValueError):
            self.constraint = OrderConstraint(lower_parameter=self.x, upper_parameter=z)


class SumConstraintTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.x = RangeParameter("x", ParameterType.INT, lower=-5, upper=5)
        self.y = RangeParameter("y", ParameterType.INT, lower=-5, upper=5)
        self.constraint1 = SumConstraint(
            parameters=[self.x, self.y], is_upper_bound=True, bound=5
        )
        self.constraint2 = SumConstraint(
            parameters=[self.x, self.y], is_upper_bound=False, bound=-5
        )

        self.constraint_repr1 = "SumConstraint(x + y <= 5.0)"
        self.constraint_repr2 = "SumConstraint(x + y >= -5.0)"

    def test_BadConstruct(self) -> None:
        with self.assertRaises(ValueError):
            SumConstraint(parameters=[self.x, self.x], is_upper_bound=False, bound=-5.0)
        z = ChoiceParameter("z", ParameterType.STRING, ["a", "b", "c"])
        with self.assertRaises(ValueError):
            # pyre-fixme[16]: `SumConstraintTest` has no attribute `constraint`.
            self.constraint = SumConstraint(
                parameters=[self.x, z], is_upper_bound=False, bound=-5.0
            )

    def test_Properties(self) -> None:
        self.assertEqual(self.constraint1.op, ComparisonOp.LEQ)
        self.assertTrue(self.constraint1._is_upper_bound)

        self.assertEqual(self.constraint2.op, ComparisonOp.GEQ)
        self.assertFalse(self.constraint2._is_upper_bound)

    def test_Repr(self) -> None:
        self.assertEqual(str(self.constraint1), self.constraint_repr1)
        self.assertEqual(str(self.constraint2), self.constraint_repr2)

    def test_Validate(self) -> None:
        self.assertTrue(self.constraint1.check({"x": 1, "y": 4}))
        self.assertTrue(self.constraint1.check({"x": 4, "y": 1}))
        self.assertFalse(self.constraint1.check({"x": 1, "y": 5}))

        self.assertTrue(self.constraint2.check({"x": -4, "y": -1}))
        self.assertTrue(self.constraint2.check({"x": -1, "y": -4}))
        self.assertFalse(self.constraint2.check({"x": -5, "y": -1}))

    def test_Clone(self) -> None:
        constraint_clone = self.constraint1.clone()
        self.assertEqual(self.constraint1.bound, constraint_clone.bound)

        constraint_clone._bound = 7.0
        self.assertNotEqual(self.constraint1.bound, constraint_clone.bound)

        constraint_clone_2 = self.constraint2.clone()
        self.assertEqual(self.constraint2.bound, constraint_clone_2.bound)

    def test_CloneWithTransformedParameters(self) -> None:
        constraint_clone = self.constraint1.clone_with_transformed_parameters(
            transformed_parameters={p.name: p for p in self.constraint1.parameters}
        )
        self.assertEqual(self.constraint1.bound, constraint_clone.bound)

        constraint_clone._bound = 7.0
        self.assertNotEqual(self.constraint1.bound, constraint_clone.bound)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Callable, Dict, Optional, Type, Union

from ax.analysis.analysis_report import AnalysisReport
from ax.analysis.base_analysis import BaseAnalysis

from ax.core.arm import Arm
from ax.core.batch_trial import AbandonedArm
from ax.core.data import Data
from ax.core.experiment import Experiment
from ax.core.generator_run import GeneratorRun, GeneratorRunType
from ax.core.metric import Metric
from ax.core.parameter import Parameter
from ax.core.parameter_constraint import ParameterConstraint
from ax.core.runner import Runner
from ax.core.trial import Trial
from ax.modelbridge.generation_strategy import GenerationStrategy
from ax.storage.json_store.registry import (
    CORE_CLASS_DECODER_REGISTRY,
    CORE_CLASS_ENCODER_REGISTRY,
    CORE_DECODER_REGISTRY,
    CORE_ENCODER_REGISTRY,
    TDecoderRegistry,
)
from ax.storage.metric_registry import CORE_METRIC_REGISTRY
from ax.storage.runner_registry import CORE_RUNNER_REGISTRY
from ax.storage.sqa_store.db import SQABase
from ax.storage.sqa_store.sqa_classes import (
    SQAAbandonedArm,
    SQAAnalysis,
    SQAAnalysisReport,
    SQAArm,
    SQAData,
    SQAExperiment,
    SQAGenerationStrategy,
    SQAGeneratorRun,
    SQAMetric,
    SQAParameter,
    SQAParameterConstraint,
    SQARunner,
    SQATrial,
)
from ax.utils.common.base import Base


@dataclass
class SQAConfig:
    """Metadata needed to save and load an experiment to SQLAlchemy.

    Attributes:
        class_to_sqa_class: Mapping of user-facing class to SQLAlchemy class
            that it will be encoded to. This allows overwriting of the default
            classes to provide custom save functionality.
        experiment_type_enum: Enum containing valid Experiment types.
        generator_run_type_enum: Enum containing valid Generator Run types.
        json_encoder_registry: Mapping from user-facing types to their json
            serialization function.
    """

    def _default_class_to_sqa_class(self=None) -> Dict[Type[Base], Type[SQABase]]:
        # pyre-ignore [7]
        return {
            AbandonedArm: SQAAbandonedArm,
            Arm: SQAArm,
            Data: SQAData,
            Experiment: SQAExperiment,
            GenerationStrategy: SQAGenerationStrategy,
            GeneratorRun: SQAGeneratorRun,
            Parameter: SQAParameter,
            ParameterConstraint: SQAParameterConstraint,
            Metric: SQAMetric,
            Runner: SQARunner,
            Trial: SQATrial,
            BaseAnalysis: SQAAnalysis,
            AnalysisReport: SQAAnalysisReport,
        }

    class_to_sqa_class: Dict[Type[Base], Type[SQABase]] = field(
        default_factory=_default_class_to_sqa_class
    )
    experiment_type_enum: Optional[Union[Enum, Type[Enum]]] = None
    generator_run_type_enum: Optional[Union[Enum, Type[Enum]]] = GeneratorRunType

    # pyre-fixme[4]: Attribute annotation cannot contain `Any`.
    # pyre-fixme[24]: Generic type `type` expects 1 type parameter, use
    #  `typing.Type` to avoid runtime subscripting errors.
    json_encoder_registry: Dict[Type, Callable[[Any], Dict[str, Any]]] = field(
        default_factory=lambda: CORE_ENCODER_REGISTRY
    )
    # pyre-fixme[4]: Attribute annotation cannot contain `Any`.
    # pyre-fixme[24]: Generic type `type` expects 1 type parameter, use
    #  `typing.Type` to avoid runtime subscripting errors.
    json_class_encoder_registry: Dict[Type, Callable[[Any], Dict[str, Any]]] = field(
        default_factory=lambda: CORE_CLASS_ENCODER_REGISTRY
    )

    json_decoder_registry: TDecoderRegistry = field(
        default_factory=lambda: CORE_DECODER_REGISTRY
    )
    # pyre-fixme[4]: Attribute annotation cannot contain `Any`.
    json_class_decoder_registry: Dict[str, Callable[[Dict[str, Any]], Any]] = field(
        default_factory=lambda: CORE_CLASS_DECODER_REGISTRY
    )

    metric_registry: Dict[Type[Metric], int] = field(
        default_factory=lambda: CORE_METRIC_REGISTRY
    )
    runner_registry: Dict[Type[Runner], int] = field(
        default_factory=lambda: CORE_RUNNER_REGISTRY
    )

    @property
    def reverse_metric_registry(self) -> Dict[int, Type[Metric]]:
        return {v: k for k, v in self.metric_registry.items()}

    @property
    def reverse_runner_registry(self) -> Dict[int, Type[Runner]]:
        return {v: k for k, v in self.runner_registry.items()}



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import argparse
import re


def patch_config(
    config_file: str, base_url: str = None, disable_algolia: bool = True
) -> None:
    config = open(config_file, "r").read()

    if base_url is not None:
        config = re.sub("baseUrl = '/';", "baseUrl = '{}';".format(base_url), config)
    if disable_algolia is True:
        config = re.sub(
            "const includeAlgolia = true;", "const includeAlgolia = false;", config
        )

    with open(config_file, "w") as outfile:
        outfile.write(config)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Path Docusaurus siteConfig.js file when building site."
    )
    parser.add_argument(
        "-f",
        "--config_file",
        metavar="path",
        required=True,
        help="Path to configuration file.",
    )
    parser.add_argument(
        "-b",
        "--base_url",
        type=str,
        required=False,
        help="Value for baseUrl.",
        default=None,
    )
    parser.add_argument(
        "--disable_algolia",
        required=False,
        action="store_true",
        help="Disable algolia.",
    )
    args = parser.parse_args()
    patch_config(args.config_file, args.base_url, args.disable_algolia)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from ax.benchmark.metrics.benchmark import BenchmarkMetric, GroundTruthBenchmarkMetric
from ax.core.arm import Arm
from ax.core.batch_trial import BatchTrial
from ax.core.trial import Trial
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_experiment


def get_test_trial() -> Trial:
    experiment = get_experiment()
    trial = experiment.new_trial()
    arm = Arm(parameters={"w": 1.0, "x": 1, "y": "foo", "z": True}, name="0_0")
    trial.add_arm(arm)
    trial.update_run_metadata(
        {
            "Ys": {"0_0": [1.0, 0.5]},
            "Ys_true": {"0_0": [1.1, 0.4]},
            "Ystds": {"0_0": [0.1, 0.1]},
            "outcome_names": ["test_metric1", "test_metric2"],
        }
    )
    return trial


def get_test_batch_trial() -> BatchTrial:
    experiment = get_experiment()
    trial = experiment.new_batch_trial()
    arm1 = Arm(parameters={"w": 1.0, "x": 1, "y": "foo", "z": True}, name="0_0")
    arm2 = Arm(parameters={"w": 1.0, "x": 2, "y": "foo", "z": True}, name="0_1")
    trial.add_arms_and_weights(arms=[arm1, arm2])
    trial.update_run_metadata(
        {
            "Ys": {"0_0": [1.0, 0.5], "0_1": [2.5, 1.5]},
            "Ys_true": {"0_0": [1.1, 0.4], "0_1": [2.5, 1.5]},
            "Ystds": {"0_0": [0.1, 0.1], "0_1": [0.0, 0.0]},
            "outcome_names": ["test_metric1", "test_metric2"],
        }
    )
    return trial


class BenchmarkMetricTest(TestCase):

    def test_fetch_trial_data(self) -> None:
        metric1 = BenchmarkMetric(name="test_metric1", lower_is_better=True)
        metric2 = BenchmarkMetric(name="test_metric2", lower_is_better=True)
        trial = get_test_trial()
        with self.assertRaisesRegex(
            NotImplementedError,
            "Arguments {'foo'} are not supported in BenchmarkMetric",
        ):
            metric1.fetch_trial_data(trial, foo="bar")
        df1 = metric1.fetch_trial_data(trial=trial).value.df  # pyre-ignore [16]
        self.assertEqual(len(df1), 1)
        self.assertDictEqual(
            df1.iloc[0].to_dict(),
            {
                "arm_name": "0_0",
                "metric_name": "test_metric1",
                "mean": 1.0,
                "sem": 0.1,
                "trial_index": 0,
            },
        )
        df2 = metric2.fetch_trial_data(trial=trial).value.df  # pyre-ignore [16]
        self.assertEqual(len(df2), 1)
        self.assertDictEqual(
            df2.iloc[0].to_dict(),
            {
                "arm_name": "0_0",
                "metric_name": "test_metric2",
                "mean": 0.5,
                "sem": 0.1,
                "trial_index": 0,
            },
        )

    def test_fetch_trial_data_batch_trial(self) -> None:
        metric1 = BenchmarkMetric(name="test_metric1", lower_is_better=True)
        metric2 = BenchmarkMetric(name="test_metric2", lower_is_better=True)
        trial = get_test_batch_trial()
        df1 = metric1.fetch_trial_data(trial=trial).value.df  # pyre-ignore [16]
        self.assertEqual(len(df1), 2)
        self.assertDictEqual(
            df1.to_dict(),
            {
                "arm_name": {0: "0_0", 1: "0_1"},
                "metric_name": {0: "test_metric1", 1: "test_metric1"},
                "mean": {0: 1.0, 1: 2.5},
                "sem": {0: 0.1, 1: 0.0},
                "trial_index": {0: 0, 1: 0},
            },
        )
        df2 = metric2.fetch_trial_data(trial=trial).value.df  # pyre-ignore [16]
        self.assertEqual(len(df2), 2)
        self.assertDictEqual(
            df2.to_dict(),
            {
                "arm_name": {0: "0_0", 1: "0_1"},
                "metric_name": {0: "test_metric2", 1: "test_metric2"},
                "mean": {0: 0.5, 1: 1.5},
                "sem": {0: 0.1, 1: 0.0},
                "trial_index": {0: 0, 1: 0},
            },
        )

    def test_make_ground_truth_metric(self) -> None:
        metric = BenchmarkMetric(name="test_metric1", lower_is_better=True)
        gt_metric = metric.make_ground_truth_metric()
        self.assertIsInstance(gt_metric, GroundTruthBenchmarkMetric)
        self.assertEqual(gt_metric.name, "test_metric1__GROUND_TRUTH")
        self.assertEqual(gt_metric.lower_is_better, metric.lower_is_better)
        self.assertFalse(gt_metric.observe_noise_sd)  # pyre-ignore [16]
        self.assertEqual(
            gt_metric.outcome_index, metric.outcome_index  # pyre-ignore [16]
        )
        self.assertIs(gt_metric.original_metric, metric)  # pyre-ignore [16]

        trial = get_test_trial()

        with self.assertRaisesRegex(
            NotImplementedError,
            "Arguments {'foo'} are not supported in GroundTruthBenchmarkMetric",
        ):
            gt_metric.fetch_trial_data(trial, foo="bar")

        df = gt_metric.fetch_trial_data(trial=trial).value.df  # pyre-ignore [16]
        self.assertEqual(len(df), 1)
        self.assertDictEqual(
            df.iloc[0].to_dict(),
            {
                "arm_name": "0_0",
                "metric_name": "test_metric1__GROUND_TRUTH",
                "mean": 1.1,
                "sem": 0.0,
                "trial_index": 0,
            },
        )

        self.assertIs(gt_metric.make_ground_truth_metric(), gt_metric)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from logging import Logger
from typing import Any, Callable, List, TypeVar

from ax.storage.sqa_store.db import SQABase
from ax.storage.sqa_store.reduced_state import GR_LARGE_MODEL_ATTRS
from ax.storage.sqa_store.sqa_classes import (
    ONLY_ONE_FIELDS,
    ONLY_ONE_METRIC_FIELDS,
    SQAMetric,
    SQAParameter,
    SQAParameterConstraint,
    SQARunner,
)
from ax.utils.common.logger import get_logger
from sqlalchemy import event
from sqlalchemy.engine import Connection
from sqlalchemy.orm.attributes import InstrumentedAttribute
from sqlalchemy.orm.base import NO_VALUE
from sqlalchemy.orm.mapper import Mapper

T = TypeVar("T")


logger: Logger = get_logger(__name__)


def listens_for_multiple(
    targets: List[InstrumentedAttribute],
    identifier: str,
    *args: Any,
    **kwargs: Any,
    # pyre-fixme[24]: Generic type `Callable` expects 2 type parameters.
) -> Callable:
    """Analogue of SQLAlchemy `listen_for`, but applies the same listening handler
    function to multiple instrumented attributes.
    """

    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[24]: Generic type `Callable` expects 2 type parameters.
    def wrapper(fn: Callable):
        for target in targets:
            event.listen(target, identifier, fn, *args, **kwargs)
        return fn

    return wrapper


# pyre-fixme[3]: Return annotation cannot be `Any`.
def consistency_exactly_one(instance: SQABase, exactly_one_fields: List[str]) -> Any:
    """Ensure that exactly one of `exactly_one_fields` has a value set."""
    values = [getattr(instance, field) is not None for field in exactly_one_fields]
    if sum(values) != 1:
        raise ValueError(
            f"{instance.__class__.__name__} must have exactly one of the following "
            f"fields set: {', '.join(exactly_one_fields)}."
        )


@listens_for_multiple(
    targets=GR_LARGE_MODEL_ATTRS,
    identifier="set",
    # `retval=True` instruct the operation ('set' on attributes in `targets`) to use
    # the return value of decorated function to set the attribute.
    retval=True,
    # `propagate=True` ensures that targets with subclasses of SQA classes used by
    # default Ax OSS encoder inherit the event listeners.
    propagate=True,
)
def do_not_set_existing_value_to_null(
    instance: SQABase, new_value: T, old_value: T, initiator_event: event.Events
) -> T:
    no_value = [None, NO_VALUE]
    if new_value in no_value and old_value not in no_value:
        logger.debug(
            f"New value for attribute is `None` or has no value, but old value "
            f"was set, so keeping the old value ({old_value})."
        )
        return old_value
    return new_value


@event.listens_for(
    SQAParameter,
    "before_insert",
)
@event.listens_for(
    SQAParameter,
    "before_update",
)
# pyre-fixme[11]: Annotation `Mapper` is not defined as a type.
def validate_parameter(mapper: Mapper, connection: Connection, target: SQABase) -> None:
    consistency_exactly_one(target, ONLY_ONE_FIELDS)


@event.listens_for(SQAParameterConstraint, "before_insert")
@event.listens_for(SQAParameterConstraint, "before_update")
def validate_parameter_constraint(
    mapper: Mapper, connection: Connection, target: SQABase
) -> None:
    consistency_exactly_one(target, ONLY_ONE_FIELDS)


@event.listens_for(SQAMetric, "before_insert")
@event.listens_for(SQAMetric, "before_update")
def validate_metric(mapper: Mapper, connection: Connection, target: SQABase) -> None:
    consistency_exactly_one(target, ONLY_ONE_FIELDS + ONLY_ONE_METRIC_FIELDS)


@event.listens_for(SQARunner, "before_insert")
@event.listens_for(SQARunner, "before_update")
def validate_runner(mapper: Mapper, connection: Connection, target: SQABase) -> None:
    consistency_exactly_one(target, ["experiment_id", "trial_id"])



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import enum
from hashlib import md5

from ax.core.formatting_utils import DataType  # noqa F401


class DomainType(enum.Enum):
    """Class for enumerating domain types."""

    FIXED: int = 0
    RANGE: int = 1
    CHOICE: int = 2
    ENVIRONMENTAL_RANGE: int = 3


class MetricIntent(enum.Enum):
    """Class for enumerating metric use types."""

    OBJECTIVE: str = "objective"
    MULTI_OBJECTIVE: str = "multi_objective"
    SCALARIZED_OBJECTIVE: str = "scalarized_objective"
    # Additional objective is not yet supported in Ax open-source.
    ADDITIONAL_OBJECTIVE: str = "additional_objective"
    OUTCOME_CONSTRAINT: str = "outcome_constraint"
    SCALARIZED_OUTCOME_CONSTRAINT: str = "scalarized_outcome_constraint"
    OBJECTIVE_THRESHOLD: str = "objective_threshold"
    TRACKING: str = "tracking"
    RISK_MEASURE: str = "risk_measure"


class ParameterConstraintType(enum.Enum):
    """Class for enumerating parameter constraint types.

    Linear constraint is base type whereas other constraint types are
    special types of linear constraints.
    """

    LINEAR: int = 0
    ORDER: int = 1
    SUM: int = 2
    DISTRIBUTION: int = 3


def stable_hash(s: str) -> int:
    """Return an integer hash of a string that is consistent across re-invocations
    of the interpreter (unlike the built-in hash, which is salted by default).

    Args:
        s (str): String to hash.
    Returns:

        int: Hash, converted to an integer.
    """
    return int(md5(s.encode("utf-8")).hexdigest(), 16)


class AnalysisType(enum.Enum):
    """Class for enumerating different experiment analysis types."""

    ANALYSIS: str = "analysis"
    PLOTLY_VISUALIZATION: str = "plotly_visualization"



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from copy import deepcopy
from functools import partial
from typing import Any, Callable, List, Optional, Union

import torch
from ax.utils.common.typeutils import checked_cast, not_none
from ax.utils.sensitivity.derivative_gp import posterior_derivative
from botorch.models.model import Model
from botorch.posteriors.gpytorch import GPyTorchPosterior
from botorch.posteriors.posterior import Posterior
from botorch.sampling.normal import SobolQMCNormalSampler
from botorch.utils.sampling import draw_sobol_samples
from botorch.utils.transforms import unnormalize
from gpytorch.distributions import MultivariateNormal


def sample_discrete_parameters(
    input_mc_samples: torch.Tensor,
    discrete_features: Union[None, List[int]],
    bounds: torch.Tensor,
    num_mc_samples: int,
) -> torch.Tensor:
    r"""Samples the input parameters uniformly at random for the discrete features.

    Args:
        input_mc_samples: The input mc samples tensor to be modified.
        discrete_features: A list of integers (or None) of indices corresponding
            to discrete features.
        bounds: The parameter bounds.
        num_mc_samples: The number of Monte Carlo grid samples.

    Returns:
        A modified input mc samples tensor.
    """
    if discrete_features is None:
        return input_mc_samples
    all_low = bounds[0, discrete_features].to(dtype=torch.int).tolist()
    all_high = (bounds[1, discrete_features]).to(dtype=torch.int).tolist()
    for i, low, high in zip(discrete_features, all_low, all_high):
        randint = partial(torch.randint, low=low, high=high + 1)
        input_mc_samples[:, i] = randint(size=torch.Size([num_mc_samples]))
    return input_mc_samples


class GpDGSMGpMean(object):

    mean_gradients: Optional[torch.Tensor] = None
    bootstrap_indices: Optional[torch.Tensor] = None
    mean_gradients_btsp: Optional[List[torch.Tensor]] = None

    def __init__(
        self,
        model: Model,
        bounds: torch.Tensor,
        derivative_gp: bool = False,
        kernel_type: Optional[str] = None,
        Y_scale: float = 1.0,
        num_mc_samples: int = 10**4,
        input_qmc: bool = False,
        dtype: torch.dtype = torch.double,
        num_bootstrap_samples: int = 1,
        discrete_features: Optional[List[int]] = None,
    ) -> None:
        r"""Computes three types of derivative based measures:
        the gradient, the gradient square and the gradient absolute measures.

        Args:
            model: A BoTorch model.
            bounds: Parameter bounds over which to evaluate model sensitivity.
            derivative_gp: If true, the derivative of the GP is used to compute
                the gradient instead of backward.
            kernel_type: Takes "rbf" or "matern", set only if `derivative_gp` is true.
            Y_scale: Scale the derivatives by this amount, to undo scaling
                done on the training data.
            num_mc_samples: The number of MonteCarlo grid samples
            input_qmc: If True, a qmc Sobol grid is use instead of uniformly random.
            dtype: Can be provided if the GP is fit to data of type `torch.float`.
            num_bootstrap_samples: If higher than 1, the method will compute the
                dgsm measure `num_bootstrap_samples` times by selecting subsamples
                from the `input_mc_samples` and return the variance and standard error
                across all computed measures.
            discrete_features: If specified, the inputs associated with the indices in
                this list are generated using an integer-valued uniform distribution,
                rather than the default (pseudo-)random continuous uniform distribution.
        """
        # pyre-fixme[4]: Attribute must be annotated.
        self.dim = checked_cast(tuple, model.train_inputs)[0].shape[-1]
        self.derivative_gp = derivative_gp
        self.kernel_type = kernel_type
        # pyre-fixme[4]: Attribute must be annotated.
        self.bootstrap = num_bootstrap_samples > 1
        # pyre-fixme[4]: Attribute must be annotated.
        self.num_bootstrap_samples = (
            num_bootstrap_samples - 1
        )  # deduct 1 because the first is meant to be the full grid
        if self.derivative_gp and (self.kernel_type is None):
            raise ValueError("Kernel type has to be specified to use derivative GP")
        self.num_mc_samples = num_mc_samples
        if input_qmc:
            # pyre-fixme[4]: Attribute must be annotated.
            self.input_mc_samples = (
                draw_sobol_samples(bounds=bounds, n=num_mc_samples, q=1, seed=1234)
                .squeeze(1)
                .to(dtype)
            )
        else:
            self.input_mc_samples = unnormalize(
                torch.rand(num_mc_samples, self.dim, dtype=dtype),
                bounds=bounds,
            )

        # uniform integral distribution for discrete features
        self.input_mc_samples = sample_discrete_parameters(
            input_mc_samples=self.input_mc_samples,
            discrete_features=discrete_features,
            bounds=bounds,
            num_mc_samples=num_mc_samples,
        )

        if self.derivative_gp:
            posterior = posterior_derivative(
                model, self.input_mc_samples, not_none(self.kernel_type)
            )
        else:
            self.input_mc_samples.requires_grad = True
            posterior = checked_cast(
                GPyTorchPosterior, model.posterior(self.input_mc_samples)
            )
        self._compute_gradient_quantities(posterior, Y_scale)

    def _compute_gradient_quantities(
        self, posterior: Union[GPyTorchPosterior, MultivariateNormal], Y_scale: float
    ) -> None:
        if self.derivative_gp:
            self.mean_gradients = checked_cast(torch.Tensor, posterior.mean) * Y_scale
        else:
            predictive_mean = posterior.mean
            torch.sum(predictive_mean).backward()
            self.mean_gradients = (
                checked_cast(torch.Tensor, self.input_mc_samples.grad) * Y_scale
            )
        if self.bootstrap:
            subset_size = 2
            self.bootstrap_indices = torch.randint(
                0, self.num_mc_samples, (self.num_bootstrap_samples, subset_size)
            )
            self.mean_gradients_btsp = [
                torch.index_select(
                    checked_cast(torch.Tensor, self.mean_gradients), 0, indices
                )
                for indices in self.bootstrap_indices
            ]

    def aggregation(
        self, transform_fun: Callable[[torch.Tensor], torch.Tensor]
    ) -> torch.Tensor:
        gradients_measure = torch.tensor(
            [
                torch.mean(transform_fun(not_none(self.mean_gradients)[:, i]))
                for i in range(self.dim)
            ]
        )
        if not (self.bootstrap):
            return gradients_measure
        else:
            gradients_measures_btsp = [gradients_measure.unsqueeze(0)]
            for b in range(self.num_bootstrap_samples):
                gradients_measures_btsp.append(
                    torch.tensor(
                        [
                            torch.mean(
                                transform_fun(
                                    not_none(self.mean_gradients_btsp)[b][:, i]
                                )
                            )
                            for i in range(self.dim)
                        ]
                    ).unsqueeze(0)
                )
            gradients_measures_btsp = torch.cat(gradients_measures_btsp, dim=0)
            return (
                torch.cat(
                    [
                        gradients_measures_btsp.mean(dim=0).unsqueeze(0),
                        gradients_measures_btsp.var(dim=0).unsqueeze(0),
                        torch.sqrt(
                            gradients_measures_btsp.var(dim=0)
                            / (self.num_bootstrap_samples + 1)
                        ).unsqueeze(0),
                    ],
                    dim=0,
                )
                .t()
                .detach()
            )

    def gradient_measure(self) -> torch.Tensor:
        r"""Computes the gradient measure:

        Returns:
            if `self.num_bootstrap_samples > 1`
                Tensor: (values, var_mc, stderr_mc) x dim
            else
                Tensor: (values) x dim
        """
        return self.aggregation(torch.as_tensor)

    def gradient_absolute_measure(self) -> torch.Tensor:
        r"""Computes the gradient absolute measure:

        Returns:
            if `self.num_bootstrap_samples > 1`
                Tensor: (values, var_mc, stderr_mc) x dim
            else
                Tensor: (values) x dim
        """
        return self.aggregation(torch.abs)

    def gradients_square_measure(self) -> torch.Tensor:
        r"""Computes the gradient square measure:

        Returns:
            if `num_bootstrap_samples > 1`
                Tensor: (values, var_mc, stderr_mc) x dim
            else
                Tensor: (values) x dim
        """
        return self.aggregation(torch.square)


class GpDGSMGpSampling(GpDGSMGpMean):

    samples_gradients: Optional[torch.Tensor] = None
    samples_gradients_btsp: Optional[List[torch.Tensor]] = None

    def __init__(
        self,
        model: Model,
        bounds: torch.Tensor,
        num_gp_samples: int,
        derivative_gp: bool = False,
        kernel_type: Optional[str] = None,
        Y_scale: float = 1.0,
        num_mc_samples: int = 10**4,
        input_qmc: bool = False,
        gp_sample_qmc: bool = False,
        dtype: torch.dtype = torch.double,
        num_bootstrap_samples: int = 1,
    ) -> None:
        r"""Computes three types of derivative based measures:
        the gradient, the gradient square and the gradient absolute measures.

        Args:
            model: A BoTorch model.
            bounds: Parameter bounds over which to evaluate model sensitivity.
            num_gp_samples: If method is "GP samples", the number of GP samples has
                to be set.
            derivative_gp: If true, the derivative of the GP is used to compute the
                gradient instead of backward.
            kernel_type: Takes "rbf" or "matern", set only if `derivative_gp` is true.
            Y_scale: Scale the derivatives by this amount, to undo scaling done on
                the training data.
            num_mc_samples: The number of Monte Carlo grid samples.
            input_qmc: If True, a qmc Sobol grid is used instead of uniformly random.
            gp_sample_qmc: If True, the posterior sampling is done using
                `SobolQMCNormalSampler`.
            dtype: Can be provided if the GP is fit to data of type `torch.float`.
            num_bootstrap_samples: If higher than 1, the method will compute the
                dgsm measure `num_bootstrap_samples` times by selecting subsamples
                from the `input_mc_samples` and return the variance and standard error
                across all computed measures.

        Returns values of gradient_measure, gradient_absolute_measure and
        gradients_square_measure change to the following:
            if `num_bootstrap_samples > 1`:
                Tensor: (values, var_gp, stderr_gp, var_mc, stderr_mc) x dim
            else
                Tensor: (values, var_gp, stderr_gp) x dim
        """
        self.num_gp_samples = num_gp_samples
        self.gp_sample_qmc = gp_sample_qmc
        self.num_mc_samples = num_mc_samples
        super().__init__(
            model=model,
            bounds=bounds,
            derivative_gp=derivative_gp,
            kernel_type=kernel_type,
            Y_scale=Y_scale,
            num_mc_samples=num_mc_samples,
            input_qmc=input_qmc,
            dtype=dtype,
            num_bootstrap_samples=num_bootstrap_samples,
        )

    def _compute_gradient_quantities(
        self, posterior: Union[Posterior, MultivariateNormal], Y_scale: float
    ) -> None:
        if self.gp_sample_qmc:
            sampler = SobolQMCNormalSampler(
                sample_shape=torch.Size([self.num_gp_samples]), seed=0
            )
            samples = sampler(posterior)
        else:
            samples = posterior.rsample(torch.Size([self.num_gp_samples]))
        if self.derivative_gp:
            self.samples_gradients = samples * Y_scale
        else:
            samples_gradients = []
            for j in range(self.num_gp_samples):
                torch.sum(samples[j]).backward(retain_graph=True)
                samples_gradients.append(
                    deepcopy(self.input_mc_samples.grad).unsqueeze(0)
                )
                self.input_mc_samples.grad.data.zero_()
            self.samples_gradients = torch.cat(samples_gradients, dim=0) * Y_scale
        if self.bootstrap:
            subset_size = 2
            self.bootstrap_indices = torch.randint(
                0, self.num_mc_samples, (self.num_bootstrap_samples, subset_size)
            )
            self.samples_gradients_btsp = []
            for j in range(self.num_gp_samples):
                not_none(self.samples_gradients_btsp).append(
                    torch.cat(
                        [
                            torch.index_select(
                                not_none(self.samples_gradients)[j], 0, indices
                            ).unsqueeze(0)
                            for indices in not_none(self.bootstrap_indices)
                        ],
                        dim=0,
                    )
                )

    def aggregation(
        self, transform_fun: Callable[[torch.Tensor], torch.Tensor]
    ) -> torch.Tensor:
        gradients_measure_list = []
        for j in range(self.num_gp_samples):
            gradients_measure_list.append(
                torch.tensor(
                    [
                        torch.mean(
                            transform_fun(not_none(self.samples_gradients)[j][:, i])
                        )
                        for i in range(self.dim)
                    ]
                ).unsqueeze(0)
            )
        gradients_measure_list = torch.cat(gradients_measure_list, dim=0)
        if not (self.bootstrap):
            gradients_measure_mean_var = []
            for i in range(self.dim):
                gradients_measure_mean_var.append(
                    torch.tensor(
                        [
                            torch.mean(gradients_measure_list[:, i]),
                            torch.var(gradients_measure_list[:, i]),
                            torch.sqrt(
                                torch.var(gradients_measure_list[:, i])
                                / self.num_gp_samples
                            ),
                        ]
                    ).unsqueeze(0)
                )
            gradients_measure_mean_var = torch.cat(gradients_measure_mean_var, dim=0)
            return gradients_measure_mean_var
        else:
            gradients_measure_list_btsp = []
            for j in range(self.num_gp_samples):
                gradients_measure_btsp = [gradients_measure_list[j].unsqueeze(0)] + [
                    torch.tensor(
                        [
                            torch.mean(
                                transform_fun(
                                    not_none(self.samples_gradients_btsp)[j][b][:, i]
                                )
                            )
                            for i in range(self.dim)
                        ]
                    ).unsqueeze(0)
                    for b in range(self.num_bootstrap_samples)
                ]
                gradients_measure_list_btsp.append(
                    torch.cat(gradients_measure_btsp, dim=0).unsqueeze(0)
                )
            gradients_measure_list_btsp = torch.cat(gradients_measure_list_btsp, dim=0)

            var_per_bootstrap = torch.var(gradients_measure_list_btsp, dim=0)
            gp_var = torch.mean(var_per_bootstrap, dim=0)
            gp_se = torch.sqrt(gp_var / self.num_gp_samples)
            var_per_gp_sample = torch.var(gradients_measure_list_btsp, dim=1)
            mc_var = torch.mean(var_per_gp_sample, dim=0)
            mc_se = torch.sqrt(mc_var / (self.num_bootstrap_samples + 1))
            total_mean = gradients_measure_list_btsp.reshape(-1, self.dim).mean(dim=0)
            gradients_measure_mean_vargp_segp_varmc_segp = torch.cat(
                [
                    torch.tensor(
                        [total_mean[i], gp_var[i], gp_se[i], mc_var[i], mc_se[i]]
                    ).unsqueeze(0)
                    for i in range(self.dim)
                ],
                dim=0,
            )
            return gradients_measure_mean_vargp_segp_varmc_segp


def compute_derivatives_from_model_list(
    model_list: List[Model],
    bounds: torch.Tensor,
    discrete_features: Optional[List[int]] = None,
    **kwargs: Any,
) -> torch.Tensor:
    """
    Computes average derivatives of a list of models on a bounded domain. Estimation
    is according to the GP posterior mean function.

    Args:
        model_list: A list of m botorch.models.model.Model types for which to compute
            the average derivative.
        bounds: A 2 x d Tensor of lower and upper bounds of the domain of the models.
        discrete_features: If specified, the inputs associated with the indices in
            this list are generated using an integer-valued uniform distribution,
            rather than the default (pseudo-)random continuous uniform distribution.
        kwargs: Passed along to GpDGSMGpMean.

    Returns:
        A (m x d) tensor of gradient measures.
    """
    indices = []
    for model in model_list:
        sens_class = GpDGSMGpMean(
            model=model, bounds=bounds, discrete_features=discrete_features, **kwargs
        )
        indices.append(sens_class.gradient_measure())
    return torch.stack(indices)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from copy import deepcopy

from ax.core.observation import ObservationFeatures
from ax.core.parameter import ChoiceParameter, ParameterType, RangeParameter
from ax.core.search_space import SearchSpace
from ax.modelbridge.transforms.search_space_to_float import SearchSpaceToFloat
from ax.utils.common.testutils import TestCase


class SearchSpaceToFloatTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.search_space = SearchSpace(
            parameters=[
                RangeParameter(
                    "a", lower=1, upper=3, parameter_type=ParameterType.FLOAT
                ),
                ChoiceParameter(
                    "b", parameter_type=ParameterType.STRING, values=["a", "b", "c"]
                ),
            ]
        )
        self.observation_features = [
            ObservationFeatures(parameters={"a": 2, "b": "a"}),
            ObservationFeatures(parameters={"a": 3, "b": "b"}),
            ObservationFeatures(parameters={"a": 3, "b": "c"}),
        ]
        self.transformed_features = [
            ObservationFeatures(parameters={"HASH_PARAM": 805305186152.0}),
            ObservationFeatures(parameters={"HASH_PARAM": 800097055771.0}),
            ObservationFeatures(parameters={"HASH_PARAM": 1551602558.0}),
        ]
        self.t = SearchSpaceToFloat()

    def test_TransformSearchSpace(self) -> None:
        ss2 = self.search_space.clone()
        ss2 = self.t.transform_search_space(ss2)
        self.assertEqual(len(ss2.parameters), 1)
        expected_parameter = RangeParameter(
            name="HASH_PARAM",
            parameter_type=ParameterType.FLOAT,
            lower=0.0,
            upper=1e12,
        )
        self.assertEqual(ss2.parameters.get("HASH_PARAM"), expected_parameter)

    def test_TransformObservationFeatures(self) -> None:
        obs_ft2 = deepcopy(self.observation_features)
        obs_ft2 = self.t.transform_observation_features(obs_ft2)
        self.assertEqual(obs_ft2, self.transformed_features)
